
/* Hive Spin Compiler DAT block parsing */
/* This code is public domain */

#include "keyword.h"
#include "symbol.h"
#include "parse.h"
#include "hsc.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>

int inDATSection = 0;

static Symbol *baseLabel = NULL;

/* Cog programs are both in hub ram and cog ram. Hub ram is byte
 * addressed and cog ram is long addressed, but when compiling
 * the two must be kept in sync. In order to keep code simple,
 * our cogptr is also using byte addressing and is scaled as
 * necessary.
 */ 
static int cogptr; /* current location in cog ram << 2 */
static int datptr; /* current location in hub ram */

static int align(int size)
{
  if( size==2 ) {
    cogptr = ((cogptr+1)&(-2));
    datptr = ((datptr+1)&(-2));
  }
  else if( size==4) {
    cogptr = ((cogptr+3)&(-4));
    datptr = ((datptr+3)&(-4));
  }
}

static int dat_file(int pass);

/* Parse a DAT block and add the constants to the symbol table.
 * a DAT line has one of four structures:
 * 1) [label | :label] ORG|RES|FIT [<expr>]
 * 2) [label | :label] <size> <expr> [ ,<expr>* ]
 * 3) [label | :label] FILE "..."
 * 4) [label | :label] [ <cond> ] <pasm op> <dst> , <src> <effect>*
 * 5) [label | :label]
 */
void ParseDATpass1()
{
  Symbol *sym;
  DATSym *dat, *base;
  int inc, count, dump, size;
  
  MustSkip(TK_DAT);
  inDATSection = 1;
  datptr = curobj->datptr;
  cogptr = 0;
  do {
    if( TestAndSkip(TK_EOL) ) continue;
    
    /* first handle the (local) label, if any */
    sym = NULL;
    if( IsId(token.type) ) {
      if( token.text[0]==':' ) {
        /* add a local label to the linked list on the base label
         * (= last global label parsed). First test for duplicates.
         */
         if( !baseLabel ) hsc_abort("no global label preceding local label");
         base = dat = (DATSym*)(baseLabel+1);
         while( (sym = dat->next_lcl) ) {
           if( !strcmp(sym->name, token.text) ) hsc_abort("duplicate local label");
           dat = (DATSym*)(sym+1);
         }
         /* if unique, allocate & add to the head of the linked list */
         sym = AllocSymbol(token.text);
         dat = MakeDATSym(sym);
         dat->next_lcl = base->next_lcl;
         base->next_lcl = sym;
      }
      else {
        /* add a new global label and make it the base for local
         * labels that might follow it.
         */
        sym = AddSymbol(token.text, &(curobj->Globals));
        dat = MakeDATSym(sym);
        dat->next_lcl = NULL;
        baseLabel = sym;
      }
      sym->type = DAT_SYMBOL;
      NextToken();
    }
    
    inc = 0;
    
    /* case 1: handle ORG, RES, FIT */
    if( TestAndSkip(TK_ORG) ) {
      align(4);
      cogptr = 0;
      if( token.type!=TK_EOL ) {
        cogptr = EvalExpr( baseLabel ) << 2;
      }
    }
    else if( TestAndSkip(TK_RES) ) {
      align(4);
      cogptr += (token.type!=TK_EOL) ? (EvalExpr( baseLabel ) << 2) : 4;
    }
    else if( TestAndSkip(TK_FIT) ) {
      int cogfit = 0x1f0;
      if( token.type!=TK_EOL ) {
        cogfit = EvalExpr( baseLabel );
      }
      if( cogptr > (cogfit<<2) ) hsc_abort("cog origin exceeds FIT limit");
    }
    
    /* case 2: handle size specifiers BYTE, WORD and LONG,
     * calculate data size but ignore values in this pass.
     * Note that strings are converted to a comma separated
     * list of integers by the lexer.
     */
    else if( IsSize(token.type) ) {
      size = 1<<(token.type-TK_BYTE);
      align(size);
      NextToken();
      if( token.type!=TK_EOL ) {
        do {
          if( IsSize(token.type) ) {
            size = 1<<(token.type - TK_BYTE);
            NextToken();
          }
          dump = TryEvalExpr( baseLabel );
          count = 1;
          if( TestAndSkip(TK_LBRACKET) ) {
            count = EvalExpr( baseLabel );
            MustSkip(TK_RBRACKET);
          }
          inc += count * size;
        } while( TestAndSkip(TK_COMMA) );
      }
    }
    
    /* case 3: handle the FILE directive */
    else if( token.type==TK_FILE ) {
      inc += dat_file(1);
      MustSkip(TK_EOL);
    }
    
    /* case 4: handle PASM, allocate one long & skip line */
    else if( IsCond(token.type) || IsPasm(token.type) ) {
      align(4);
      inc = 4;
      size = 4;
      do { NextToken(); } while( token.type!=TK_EOL );
    }
    
    /* case 5: a label on a line by itself */
    else if( token.type!=TK_EOL ) hsc_abort("syntax error");
    
    /* now set the label (if any) data correctly */
    if( sym ) {
      sym->size = size;
      dat->hubloc = datptr;
      dat->cogloc = cogptr;
    }
    datptr += inc;
    cogptr += inc;

  } while( !IsBlockDesignator(token.type) && token.type!=TK_EOF );
  curobj->datptr = datptr;
  inDATSection = 0;
}

static void emit_byte(int d)
{
//  printf("%02x ", d&0xff);
  EmitDATByte(d); /* increments curobj->datptr */
  ++datptr;
  ++cogptr;
}

#define HAS_SRC   1
#define HAS_DST   2
#define ALWAYS    0xf
#define NOP       0

void ParseDATpass2()
{
  int val, tmp, i, count, size;
  int opcode, cond, effects, ds;
  Symbol *sym;
  DATSym *dat;
  
  MustSkip(TK_DAT);
  inDATSection = 1;
  datptr = curobj->datptr - curobj->pbase;
  cogptr = 0;
  do {
    if( TestAndSkip(TK_EOL) ) continue;
//    printf("\n(%04d,%02d) 0x%04x (0x%04x): ", token.line, token.col, datptr, cogptr>>2);
    
    /* first locate the label, if any */
    sym = NULL;
    dat = NULL;
    if( IsId(token.type) ) {
      if( token.text[0]==':' ) {
         dat = (DATSym*)(baseLabel+1);
         while( (sym = dat->next_lcl) ) {
           dat = (DATSym*)(sym+1);
           if( !strcmp(sym->name, token.text) ) break;
         }         
      }
      else {
        sym = SymbolLookup(token.text, curobj->Globals);
        if( sym->type==DAT_SYMBOL ) dat = (DATSym*)(sym+1);
        baseLabel = sym;
      }
      NextToken();
    }

    /* case 1: handle ORG, RES, FIT */
    if( TestAndSkip(TK_ORG) ) {
      align(4);
      cogptr = 0;
      if( token.type!=TK_EOL ) {
        cogptr = EvalExpr( baseLabel ) << 2;
      }
    }
    else if( TestAndSkip(TK_RES) ) {
      align(4);
      cogptr += (token.type!=TK_EOL) ? (EvalExpr( baseLabel ) << 2) : 4;
    }
    else if( TestAndSkip(TK_FIT) ) {
      /* ignore FIT in second pass */
      do { NextToken(); } while( token.type!=TK_EOL );      
    }

#if 1
    if( sym && dat->hubloc!=datptr ) hsc_abort("internal error: datptr mismatch");
#endif
    
    /* case 2: handle size specifiers BYTE, WORD and LONG */
    else if( IsSize(token.type) ) {
      size = 1<<(token.type-TK_BYTE);
      align(size);
      NextToken();
      if( token.type!=TK_EOL ) {
        do {
          if( IsSize(token.type) ) {
            size = 1<<(token.type - TK_BYTE);
            NextToken();
          }      
          val = EvalExpr( baseLabel );
          count = 1;
          if( TestAndSkip(TK_LBRACKET) ) {
            count = EvalExpr( baseLabel );
            MustSkip(TK_RBRACKET);
          }
          while( count-- ) {
            int tmp = val;
            for(i=0; i<size; ++i) {
              emit_byte( tmp );
              tmp >>= 8;
            }
          }
        } while( TestAndSkip(TK_COMMA) );
      }
    }
    
    /* case 3: handle the FILE directive */
    else if( token.type==TK_FILE ) {
      dat_file(2);
      MustSkip(TK_EOL);
    }
        
    /* case 4: handle PASM, this pass we do the assembly */
    else if( IsCond(token.type) || IsPasm(token.type) ) {
      align(4);
      
      /* first we check & handle a condition code preceding the instruction */
      cond = ALWAYS;
      if( IsCond(token.type) ) {
        cond = GetCond(token.type);
        NextToken();
      }
      
      if( !IsPasm(token.type) ) hsc_abort("expected PASM instruction");
      ds = HAS_DST | HAS_SRC;
      /* some tokens are both Spin and PASM and these are returned by
       * the lexer as Spin tokens, hence we special case those here.
       */
      switch( token.type ) {
        case TK_AND:        opcode = 0x60bc0000; break;
        case TK_OR:         opcode = 0x68bc0000; break;
        case TK_WAITCNT:    opcode = 0xf8bc0000; break;
        case TK_WAITPEQ:    opcode = 0xf03c0000; break;
        case TK_WAITPNE:    opcode = 0xf43c0000; break;
        case TK_WAITVID:    opcode = 0x0cfc0001; break;
        case TK_COGID:      opcode = 0x0cfc0001; ds = HAS_DST; break;
        case TK_COGINIT:    opcode = 0x0c7c0002; ds = HAS_DST; break;
        case TK_COGSTOP:    opcode = 0x0cfc0003; ds = HAS_DST; break;
        case TK_CLKSET:     opcode = 0x0c7c0000; ds = HAS_DST; break;
        
        default:            opcode = GetPasmOp(token.type);
                            ds = GetPasmDS(token.type);
      }
 
      /* the CALL instruction requires special handling of the call
       * and return labels.
       */  
      if( token.type==TK_CALL ) {
        char buf[MAXTOKENLENGTH+1];
        
        NextToken();
        MustSkip(TK_HASH);
        if( token.type!=TK_ID ) hsc_abort("expected call label");
        
        /* find and check the return label */
        strcpy(buf, token.text);
        strcat(buf, "_RET");
        if( !(sym = SymbolLookup(buf, curobj->Globals)) ) hsc_abort("return label not found");
        if( sym->type!=DAT_SYMBOL ) hsc_abort("return label is not in DAT block");
        dat = (DATSym*)(sym+1);
        if( dat->cogloc & 3 ) hsc_abort("return label is not long aligned");
        
        /* or in the call destination and return locations */
        opcode |= ((dat->cogloc>>2) & 0x1ff) << 9;
        opcode |= EvalExpr(baseLabel) & 0x1ff;
      }
      
      /* all other instructions: handle the operands as needed */
      else {
        NextToken();
        if( ds & HAS_DST ) opcode |= (EvalExpr(baseLabel) & 0x1ff)<<9;
        if( ds == (HAS_DST|HAS_SRC) ) MustSkip(TK_COMMA);
        if( ds & HAS_SRC ) {
          if( TestAndSkip(TK_HASH) ) opcode |= 1<<22;
          opcode |= (EvalExpr(baseLabel) & 0x1ff);
        }
      }
    
      /* effects are 3 bits: WC, WZ and WR; pseudo-effect NR resets the WR bit */
      effects = 0;
      if( token.type!=TK_EOL ) {
        do {
          if( !IsEffect(token.type) ) hsc_abort("expected PASM effect");
          if( GetEffect(token.type)==TK_NR ) {
            effects &= TK_WC|TK_WZ ;
          }
          else {
            effects |= GetEffect(token.type);
          }
          NextToken();
        } while( TestAndSkip(TK_COMMA) );
      }

      /* fold in conditions and effects, and emit the opcode */
      if( opcode==NOP && (cond!=ALWAYS || effects) )
        hsc_abort("no conditions or effects possible on NOP");
      opcode = (opcode & 0xffc3ffff) | (cond<<18);
      opcode |= effects << 23;
      for(i=0; i<4; ++i) {
        emit_byte( opcode );
        opcode >>= 8;
      }
    }
    
    /* case 5: nothing left to do */
    
  } while( !IsBlockDesignator(token.type) && token.type!=TK_EOF );
  assert( curobj->datptr==datptr+curobj->pbase );
  inDATSection = 0;
}

#define MAXPATH 4096
static char path[MAXPATH];

static int dat_file(int pass)
{
  FILE *f;
  int c, i = 0;
  
  NextToken();
  do {
    if( !IsIntLiteral(token.type) ) hsc_abort("syntax error");
    path[i++] = token.value;
    NextToken();
  } while( TestAndSkip(TK_COMMA) );
  path[i] = 0;
  
  i = 0;
  f = fopen(path, "rb");
  if( !f ) hsc_abort("could not open file");
  while( (c = fgetc(f))!=EOF ) {
    if( pass==2 ) emit_byte(c);
    ++i;
  }
  fclose(f);
  
  return i;
}
