
/* Hive Spin Compiler constant expression handling */
/* This code is public domain */

#include "symbol.h"
#include "keyword.h"
#include "opcodes.h"
#include "parse.h"
#include "hsc.h"

#define LABEL_ADDR 1
#define SYMBOL_VAL 0

#define MIN_PRIO  13

static Symbol *localsyms;
static int eval(int);
static int ignore_undef;

static int eval(int rhp);
static int eval_symbol(int mode);
static int eval_unary(int type);
static int eval_binary(int type, int left, int right);

int evalOK;

int TryEvalExpr(Symbol *table)
{
  localsyms = table;
  ignore_undef = 1;
  evalOK = 1;
  return eval( MIN_PRIO );
}

int EvalExpr(Symbol *table)
{
  localsyms = table;
  ignore_undef = 0;
  evalOK = 1;
  return eval( MIN_PRIO );
}

/* Evaluate a constant expression. Algorithm uses a simple bottom-up
 * expression parser using Dijkstra's shunting algorithm. Operator/
 * operand stack is implemented with recursion. Expression is
 * calculated as it is being parsed. Returns the value of the
 * expression.
 */
static int eval(int rhp)
{
  int val, lhp, t = token.type;
  
  /* first evaluate the first operand; note that it may start with
   * an unary prefix operator or could be a subexpression.
   */
  if( (IsUnaryOp(t) && !IsPostfixOp(t)) || t==TK_MINUS ) {
    NextToken();
    val = eval_unary(t);
  }
  else if( IsReg(t) ) {
    val = ( 0x1f0+GetReg(t) );
    NextToken();
  }
  else switch( t ) {
    case TK_AT:         NextToken();
                        val = eval_symbol(LABEL_ADDR);
                        break;
                        
    case TK_LPAREN:     NextToken();
                        val = eval( MIN_PRIO );
                        MustSkip( TK_RPAREN );
                        break;
                        
    case TK_INTLITERAL: val = token.value;
                        NextToken();
                        break;

    case TK_FLOATLITERAL: val = token.fltval;
                        NextToken();
                        break;
                        
    default:            val = eval_symbol(SYMBOL_VAL);
  }

  /* At this point val holds the value of the operand, now check
   * for an operator. Note that we do not handle postfix operators
   * because they all operate on variables which have no place in
   * a constant expression.
   */
  while( 1 ) {
    /* Note that GetPrecedence returns MIN_PRIO for non-operators */
    t = token.type;
    lhp = GetPrecedence(t);
    /* all operators are now left associative: correct? (use rhp<lhp
     * for right associative)
     */
    if( rhp<=lhp ) break;
    NextToken();
    val = eval_binary(t, val, eval(lhp));
  }
  return val;
}

/* resolve a symbol to its value. Depending on the type of symbol and
 * on the mode the meaning of value differs. 'mode' can be LABEL_ADDR, in
 * which case its address of the symbol is returned, or SYMBOL_VAL, in
 * which case its value is returned.
 */
static int eval_symbol(int mode)
{
  Symbol *p;
  DATSym *q;
  OBJSym *obj;
  int val;
  
  /* first we handle local labels, which are only valid in a DAT section,
   * following a global symbol. This global symbol (stored in 'localsyms')
   * is the base for a linked list of the local symbols following it.
   */
  if( token.text[0]==':') {
    if( !localsyms ) hsc_abort("local label unknown in current context");
    p = localsyms;
    while(1) {
      /* p points to a DAT symbol, which has a DATSym struct
       * immediately following it.
       */
      q = (DATSym*)(p+1);
      if( !(p = q->next_lcl) ) {
        if( ignore_undef ) {
          evalOK = 0;
          NextToken();
          return -1;
        }
        hsc_abort("local label unknown in current context");
      }
      if( !strcmp(p->name, token.text) ) break;
    }
  }
  /* next we handle regular symbols, which is more straightforward */
  else {
    if( !(p = SymbolLookup(token.text, curobj->Globals)) ) {
      if( ignore_undef ) {
        evalOK = 0;
        NextToken();
        return -1;
      }
      hsc_abort("unknown symbol");
    }
  }
  
  /* at this point p points to the right symbol */
  switch( p->type ) {
    
    case INT_CON_SYMBOL:
    case FLT_CON_SYMBOL:
      if( mode==LABEL_ADDR ) hsc_abort("cannot take address of literal");
      val = p->value;
      NextToken();
      return val;
      
    case UNDEF_CON_SYMBOL:
      if( ignore_undef ) {
        evalOK = 0;
        NextToken();
        return -1;
      }
      hsc_abort("undefined CON symbol");
      
    case OBJ_SYMBOL:
      /* in a constant expression, an OBJ symbol must be followed by #id,
       * where id is the name of a constant defined in the object.
       */
      if( mode==LABEL_ADDR ) hsc_abort("cannot take address of OBJ constant");
      obj = (OBJSym *)(p+1);
      MustSkip(TK_HASH);
      p = SymbolLookup( token.text, obj->sym_table);
      if( !p ) hsc_abort("constant unknown in imported object");
      val = p->value;
      NextToken();
      return val;
      
    case DAT_SYMBOL:
      /* during the first pass, we ignore DAT constants */
      if( ignore_undef ) {
        evalOK = 0;
        NextToken();
        return -1;
      }
      /* outside a DAT section, or when using the @ operator the
       * label refers to the hubram address of the symbol. Inside
       * a DAT section it refers to the cogram address.
       */
      q = (DATSym*)(p+1);
      if( !inDATSection || mode==LABEL_ADDR ) {
        val = q->hubloc;
        NextToken();
        return val;
      }
      val = q->cogloc;
      if( val & 3 ) hsc_abort("cog address not long");
      val >>= 2;
      NextToken();
      return val;
  }
}

/* Algorithm taken from spinsim */
static int intsqrt(int val)
{
  int tmp2 = 0, tmp3 = 1<<30;
  
  while( tmp3 ) {
    tmp2 |= tmp3;
    if( tmp2 <= val) {
      val  -= tmp2;
      tmp2 += tmp3;
    }
    else {
      tmp2 -= tmp3;
    }
    tmp2 >>= 1;
    tmp3 >>= 2;
  }
  return tmp2;
}

static int eval_unary(int type)
{
  int i, val = eval( GetPrecedence(type) );
  switch( type & 0xff ) {
    case OP_MINUS:  return (-val);
    case OP_BITNOT: return (~val);
    case OP_NOT:    return (val?0:-1);
    case OP_SQRT:   return intsqrt(val);
    case OP_ABS:    return (val<0?-val:val);
    case OP_DECODE: return (1<<val);
    case OP_ENCODE: 
      for(i=32; i>=1; --i) {
	      if( val & 0x80000000 ) return i;
	      val <<= 1;
	    }
    default: hsc_abort("syntax error in constant expression");    
  }
}

static int bitrev(int val, int nbits)
{
  int retval = 0;

  while( nbits-- > 0 )  {
    retval = (retval << 1) | (val & 1);
    val >>= 1;
  }
  return retval;
}

#include "stdint.h"

static int eval_binary(int type, int left, int right)
{
  int32_t a = left;
  int32_t b = right;
  
  switch( type & 0xff ) {
  case OP_PLUS:     return (a + b);
  case OP_MINUS:    return (a - b);
  case OP_MUL:      return (a * b);
  case OP_MULH: {
    int64_t a1 = left;
    int64_t b1 = right;
    return (a1*b1)>>32;
    }
  case OP_DIV:      return (a / b);
  case OP_MOD:      return (b ? a%b : 0);
  case OP_ROR:      return ((a>>b) | (a<<(32-b)));
  case OP_ROL:      return ((a>>(32-b)) | (a<<b));
  case OP_SHR:      return (((uint32_t)a) >> b);
  case OP_SHL:      return (a << b);
  case OP_SAR:      return (a >> b);
  case OP_REV:      return bitrev(a, b);
  case OP_BITAND:   return (a & b);
  case OP_BITOR:    return (a | b);
  case OP_BITXOR:   return (a ^ b);
  case OP_MIN:      return (a>b ? b : a);
  case OP_MAX:      return (a>b ? a : b);
  case OP_LT:       return (a < b);
  case OP_GT:       return (a > b);
  case OP_EQ:       return (a == b);
  case OP_NE:       return (a != b);
  case OP_LTEQ:     return (a <= b);
  case OP_GTEQ:     return (a >= b);
  case OP_AND:      return ((a&&b) ? -1 : 0);
  case OP_OR:       return ((a||b) ? -1 : 0);
  default: hsc_abort("syntax error in constant expression");
  }
}
