
/* Hive Spin Compiler expression parser */
/* This code is public domain */

#include "symbol.h"
#include "keyword.h"
#include "parse.h"
#include "hsc.h"
#include <assert.h>

#define NULL ((void*)0)
#define EVOID      0
#define EPUSH      1
#define MIN_PRIO  13

static Node* parse_expr2(int rhp);
static Node* parse_primary(int mode);

static Node* new_intnode(int val)
{
  Node *n = new_node(TK_INTLITERAL, NULL);
  n->val = val;
  return n;  
}

Node* ParseExpr()
{
  return parse_expr2(MIN_PRIO);
}

Node* ParseVoidExpr()
{
  Node *n = parse_primary(EVOID);
  return new_node(TK_VOID, n);
}

static Symbol* lookup_id()
{
  Symbol *sym = SymbolLookup(token.text, Locals);
  if( !sym ) sym = SymbolLookup(token.text, curobj->Globals);
  if( !sym ) hsc_abort("unknown identifier");
  return sym;
}

Node* parse_arglist(int argc)
{
  int cnt = 0;
  Node *al, *n;
  
  if( !argc ) return NULL;
  MustSkip(TK_LPAREN);
  al = n = new_node(TK_LIST, NULL);
  do {
    ++cnt;
    n->sub = ParseExpr();
    
    /* special handling for LOOKUP, etc. arg lists */
    if( argc==-1) {
      if( cnt==1 ) {
        if( token.type!=TK_COLON) hsc_abort("expected colon");
        token.type = TK_COMMA;
      }
      else if( TestAndSkip(TK_DOTDOT) ) {
        n->sub = new_node2(TK_RANGE, n->sub, ParseExpr());
      }
    }
    
    if( token.type==TK_COMMA ) n->nxt = new_node(TK_LIST, NULL);
    n = n->nxt;
  } while( TestAndSkip(TK_COMMA) );
  MustSkip(TK_RPAREN);
  if( argc!=-1 && argc!=cnt ) hsc_abort("wrong number of arguments");
  return al;
}

/* Operator        
 * precendence:    Associativity:     Operator:
 * 
 * Highest   0     right-to-left      --,++,~,~~,?, @,@@       (prefix)
 *           0     left-to-right      --,++,~,~~,?             (suffix)
 *           1     right-to-left      +, -, ^^, ||, |<, >|, !
 *           2     left-to-right      ->, <-, >>, <<, ~>, ><
 *           3     left-to-right      &
 *           4     left-to-right      |, ^
 *           5     left-to-right      *, **, /, //
 *           6     left-to-right      +, -
 *           7     left-to-right      #>, <#
 *           8     left-to-right      <, >, <>, ==, =<, =>
 *           9     right-to-left      NOT
 *          10     left-to-right      AND
 *          11     left-to-right      OR
 * Lowest   12     right-to-left      =, :=, all assignment operators
 * MIN_PRIO 13                        non-expression tokens
 *
 */

/*
 * expr ::=
 *   <primary>
 *   <expr> binop <primary>
 *
 * Parse an expression. It uses Keith Clark's operator precedence
 * climbing algorithm, which is a recursive variant of Dijkstra's
 * shunting yard algorithm.
 * http://en.wikipedia.org/wiki/Operator-precedence_parser
 */
static Node* parse_expr2(int lhp)
{
  int rhp, t;
  Node *expr;
  
  expr = parse_primary(EPUSH);

  t = token.type;
  while( IsBinaryOp(t) ) {
    rhp = GetPrecedence(t);
    /* all non-assignment binary operators are left associative */
    if( lhp<=rhp ) break;
    NextToken();
    expr = new_node2(t, expr, parse_expr2(rhp));
    t = token.type;
  }
  return expr;
}

/*
primary ::=
  (<expr>)
  @@ <expr>
  @ <lval>
  prefixop <lval>
  unaryop <primary>
  <lval> suffixop
  <lval> assignop <expr>
  <lval>
  <factor>
  <builtin>
  <constant>
*/
static Node* parse_lval(int prefix, Symbol *sym);
static Node* parse_factor(Symbol *sym, int mode);
static Node* parse_builtin(int mode);
static Node* parse_constant();

#define PRIO_UMINUS 1
#define PRIO_UPLUS  1

static Node* parse_primary(int mode)
{
  int cls, t, val;
  Symbol *sym = NULL;
  Node *n;
  
  /* for ease of parsing, group various operators in a class; identifiers
   * are replaced by their symbol type and a reference to the symbol.
   */
  cls = t  = token.type;
  if( IsUnaryOp(t) )    cls = TK_UNARY;
  if( IsPostfixOp(t) )  cls = TK_PREFIX; /* prefix and suffix are the same set */
  if( IsBuiltin(t) )    cls = TK_BUILTIN;
  
  /* check mode validity of operators that are handled inline */
  if( (t==TK_LPAREN || t==TK_AT || t==TK_ATAT) && mode==EVOID ) {
    hsc_abort("not valid in void expression");
  }
  
  switch( cls ) {
    case TK_MINUS:
    case TK_UNARY:
      if( mode==EVOID ) {
        NextToken();
        return new_node(t, parse_lval(t, NULL));
      }
      /* fall through */
    case TK_LPAREN:
    case TK_PLUS:
    case TK_ATAT:
    case TK_AT:
    case TK_PREFIX:
      NextToken();
  }
  
  if( t==TK_ID || t==TK_RESULT ) {
    sym = lookup_id();
    cls = t = sym->type;
  }

  switch( cls ) {
    case TK_LPAREN:
      n = ParseExpr();
      MustSkip(TK_RPAREN);
      return n;

    case TK_ATAT:         return new_node(t, ParseExpr());

    /* note that '+' and '-' are coded as binary opertors by the lexer,
     * hence we need special handling here; unary '+' is ignored, and
     * unary '-' gets a TK_UMINUS node.
     */ 
    case TK_PLUS:         return parse_expr2(PRIO_UPLUS);
    case TK_MINUS:        return new_node(TK_UMINUS, parse_expr2(PRIO_UMINUS) );
    case TK_UNARY:        return new_node(t,         parse_expr2(GetPrecedence(t)));

    case TK_AT:
    case TK_PREFIX:       return new_node(t, parse_lval(t, NULL));

    case TK_BUILTIN:      return parse_builtin(mode);

    case TK_BACKSLASH:
    case PUB_SYMBOL:
    case PRI_SYMBOL: 
    case OBJ_SYMBOL:
    case INT_CON_SYMBOL:
    case FLT_CON_SYMBOL:  return parse_factor(sym, mode);

    case TK_INTLITERAL:
    case TK_FLOATLITERAL:
    case TK_CONSTANT:
    case TK_STRING:
      if( mode==EVOID ) hsc_abort("not valid in void expression");
      return parse_constant();

    default:
      n = parse_lval(0, sym);
      t = token.type;
      if( IsPostfixOp(t) ) {
        NextToken();
        return new_node2(t, NULL, n);
      }
      if( IsAssignmentOp(t) ) {
        NextToken();
        return new_node2(t, n, parse_expr2(12));
      }
      if( mode==EVOID ) hsc_abort("not valid in void expression");
  }
  return n;
}

/*
lval ::=
  dat_symbol
  var_symbol
  lcl_symbol
  SPR[<expr>]
  <mem>[<expr>]
  <mem>[<expr>][<expr>]
  <reg>
  <reg>[<expr>]
  <reg>[<expr>..<expr>]

mem ::= LONG | WORD | BYTE
reg ::= PAR | CNT | INA | ... etc.
*/
static Node *parse_lval(int prefix, Symbol *sym)
{
  int t, cls;
  Node* n;

  /* for ease of parsing, group various operators in a class; identifiers
   * are replaced by their symbol type and a reference to the symbol.
   * If the lval follows a prefix or @ op, it must check for an identifier
   * itself; in all other cases it gets sym passed in.
   */
  t = token.type;
  if( sym ) {
    t = sym->type;
  }
  else if( token.type==TK_ID || token.type==TK_RESULT ) {
    sym = lookup_id();
    t = sym->type;
  }
  cls = t;
  if( IsSize(t) ) cls = TK_MEM;
  if( IsReg(t) )  cls = TK_REG;
  NextToken();
  
  /* parse the lval by class */
  switch( cls ) {
    case DAT_SYMBOL:
    case VAR_SYMBOL:
    case LCL_SYMBOL:
      n = new_node(t, (Node*)sym );
      if( TestAndSkip(TK_LBRACKET) ) {
        n->nxt = ParseExpr();
        MustSkip(TK_RBRACKET);
      }
      return n;

    case TK_MEM:
      MustSkip(TK_LBRACKET);
      n = new_node(t, ParseExpr());
      MustSkip(TK_RBRACKET);
      if( TestAndSkip(TK_LBRACKET) ) {
        n->nxt = ParseExpr();
        MustSkip(TK_RBRACKET);
      }
      return n;
      
    case TK_SPR:
      if( prefix==TK_AT ) hsc_abort("cannot take address of register");
      MustSkip(TK_LBRACKET);
      n = new_node(t, ParseExpr());
      MustSkip(TK_RBRACKET);
      return n;
        
    case TK_REG:
      if( prefix==TK_AT ) hsc_abort("cannot take address of register");
      n = new_node(t, NULL);
      if( TestAndSkip(TK_LBRACKET) ) {
        n->sub = ParseExpr();
        if( TestAndSkip(TK_DOTDOT) ) {
          n->nxt = ParseExpr();
        }
        MustSkip(TK_RBRACKET);
      }
      return n;
    
    default:
      hsc_abort("syntax error in expression");
  }
}

/*
 * factor ::=
 *   <constant_id>
 *   \<method_call>
 *   <method_call>
 * 
 * method_call ::=
 *   method<arg_list>
 *   obj.method<arg_list>
 *   obj[<expr>].method<arg_list>
 * 
 * constant_id ::=
 *   id
 *   obj#id
 */
static Node* parse_factor(Symbol *sym, int mode)
{
  int t, op, abortcatch = 0;
  Node *n, *m, *args, *idx = NULL;
  Symbol *obj = NULL;
  
  t = sym ? sym->type : token.type;
  NextToken();
  if( t==TK_BACKSLASH ) {
    t = token.type;
    if( t!=TK_ID && t!=TK_RESULT ) hsc_abort("expected identifier");
    sym = lookup_id();
    t = sym->type;
    abortcatch = 1;
    NextToken();
  }
  
  switch( t ) {
    case OBJ_SYMBOL:
      if( TestAndSkip(TK_LBRACKET) ) {
        idx = ParseExpr();
        MustSkip(TK_RBRACKET);
      }
      if( (t = token.type)!=TK_DOT && t!=TK_HASH ) hsc_abort("expect . or #");
      NextToken();
      if( token.type!=TK_ID ) hsc_abort("expected identifier");
      obj = sym;
      sym = SymbolLookup(token.text, sym->obj->Globals);
      if( !sym ) hsc_abort("unknown identifier");
      NextToken();
      if( t==TK_DOT ) break;

      if( sym->type!=INT_CON_SYMBOL && sym->type!=FLT_CON_SYMBOL ) {
        hsc_abort("expected object constant");
      }
      /* fall through */
    case INT_CON_SYMBOL:
    case FLT_CON_SYMBOL:
      if( mode==EVOID ) hsc_abort("not valid in void expression");
      if( abortcatch ) hsc_abort("expected method");
      return new_intnode(sym->value);
  }
  
  /* Once we are here it must be method call.  For a simple method call 
   * we have one node of type TK_FRAMEx, with the method symbol in 'sub'
   * and the arg list in 'nxt'. For an object-method call there is a top
   * node of type OBJ_SYMBOL with 'sub' holding a node as above and 'nxt'
   * holding a TK_LIST node with the object symbol in 'sub' and the index
   * expression, if any, in 'nxt'.
   */
  t = sym->type;
  if( t!=PUB_SYMBOL && t!=PRI_SYMBOL ) hsc_abort("expected method");
  op = TK_FRAME0 + (abortcatch ? 2 : 0) + ((mode==EPUSH) ? 0 : 1);
  args = parse_arglist(sym->argc);
  n = new_node2(op, (Node*)sym, args);
  if( obj ) {
    m = new_node2(TK_LIST, (Node*)obj, idx);
    n = new_node2(OBJ_SYMBOL, n, m);
  }
  return n;
}

/* Parsing builtin functions is straightforward, just build a node with
 * an argument list attached a when needed. Biggest complication is deciding
 * whether the builtin function can, may or cannot appear in a void expression.
 */
static Node* parse_builtin(int mode)
{
  Node *m, *n = NULL;
  int t = token.type;
  int argc = GetNumArgs(t);

  NextToken();
  switch( t ) {
    case TK_BYTEFILL:      case TK_WORDFILL:      case TK_LONGFILL:  
    case TK_BYTEMOVE:      case TK_WORDMOVE:      case TK_LONGMOVE:
    case TK_WAITPEQ:       case TK_WAITPNE:       case TK_CLKSET:    
    case TK_COGSTOP:       case TK_LOCKRET:       case TK_WAITCNT:   
    case TK_WAITVID:       case TK_REBOOT:
      if( mode!=EVOID ) hsc_abort("not valid in expression");
      break;
 
    case TK_COGID:         case TK_CHIPVER:       case TK_CLKFREQ:   
    case TK_CLKMODE:       case TK_STRCOMP:       case TK_STRSIZE:   
    case TK_LOOKDOWN:      case TK_LOOKDOWNZ:     case TK_LOOKUP:    
    case TK_LOOKUPZ:
      if( mode!=EPUSH ) hsc_abort("not valid in void expression");
  }

  if( argc==0xf ) argc = -1;
  if( argc ) n = parse_arglist(argc);

  /* COGNEW(a, b) is just sugar for COGINIT(-1, a, b) */
  if( t==TK_COGNEW ) {
    m = new_node(TK_INTLITERAL, (Node*)(-1));
    n = new_node2(TK_LIST, m, n);
    t = TK_COGINIT;
  }
  
  return new_node(t, n);
}

/* constant ::=
 *   <int literal> | <flt literal>
 *   CONSTANT( <expr> )
 *   STRING("...")
 * 
 * Note that strings are converted to INT, INT, ... sequences by the lexer.
 */
static Node* parse_constant()
{
  Node *expr;
  int t = token.type, val = token.value;
  
  NextToken();
  switch( t ) {
    case TK_INTLITERAL:
    case TK_FLOATLITERAL:
      return new_intnode(val);
      
    case TK_CONSTANT:
      MustSkip(TK_LPAREN);
      val = EvalExpr(NULL);
      MustSkip(TK_RPAREN);
      return new_intnode(val);
      
    case TK_STRING:
//      expr = new_intnode(curobj->datptr - curobj->pbase);
      expr = new_intnode(curobj->datptr - 0);
      MustSkip(TK_LPAREN);
      do {
        if( token.type!=TK_INTLITERAL ) hsc_abort("not a string or byte sequence");
        EmitDATByte( token.value );
        NextToken();
      } while( TestAndSkip(TK_COMMA) );
      EmitDATByte(0);
      MustSkip(TK_RPAREN);
      return expr;  
  }
  hsc_abort("internal: bad case in parse_constant");
  return 0;
}
