
/* Hive Spin Compiler PNUT object linker */
/* This code is public domain */

#include "symbol.h"
#include "keyword.h"
#include "parse.h"
#include "hsc.h"
#include <stdio.h>
#include <string.h>
#include <assert.h>

static unsigned char hubram[32768];
int hubptr;

struct _pgmhdr {
  int   clkfreq;
  char  clkmode;
  char  checksum;
  short pbase;
  short vbase;
  short dbase;
  short pcurr;
  short dcurr;
};
typedef struct _pgmhdr PgmHdr;


/* ObjHdr is the single long at the start of a object */
struct _objhdr {
  short objlen; /* length of object in bytes      */
  char  obj1st; /* index of first imported object */
  char  objcnt; /* count of imported objects      */
};
typedef struct _objhdr ObjHdr;

/* PubLnk is call information for each method in the current object. For a
 * method with index 'i', the link information is at LONG[PBASE][i].
 */
struct _publnk {
  short paddr;  /* address of the PUB/PRI method    */
  short lclsiz; /* size of local variables in bytes */
};
typedef struct _publnk PubLnk;

/* ObjLnk is call information for method calls in imported objects. To locate
 * an imported object instance VBASE is increased by 'vdiff' and PBASE is 
 * increased by 'pdiff' (note that pdiff can be signed). For imported object
 * 'i' the link information is at LONG[PBASE][i].
 */
struct _objlnk {
  union {
    short pdiff; /* difference from current PBASE to imported object PBASE   */
    short idx;   /* object index -- placeholder during linking               */
  };
  short vdiff;   /* difference from current VBASE to imported instance VBASE */
};
typedef struct _objlnk ObjLnk;

extern int pnutpass;

void EmitByte(int byte)
{
  unsigned char c = byte & 0xff;
  if( pnutpass==2 ) hubram[hubptr] = c;
  ++hubptr;
  if( hubptr>0x7000 ) hsc_abort("program too large");
}

void EmitDATByte(int byte)
{
  unsigned char c = byte & 0xff;
  hubram[curobj->datptr++] = c;
  if( curobj->datptr>0x7000 ) hsc_abort("program too large");
}

void EmitObjectHdr(Object *obj)
{
  ObjHdr *objhdr;

  if( pass==1 ) {
    obj->vcurr = 0;
    obj->datptr = 0;
    return;
  }
  
  printf("\nObject %s, pbase = %04x\n", obj->path, hubptr);
  
  obj->pbase = hubptr;

  /* fill out the object header structure (1 long) */
  objhdr = (ObjHdr*)(hubram + hubptr);  
  objhdr->obj1st = 1 + obj->nPubs-1 + obj->nPris-1;
  objhdr->objcnt = obj->nImps-1;

  /* allocate space for above header long, for the link entries for the PUB
   * and PRI symbols, and for the imported objects (one long each).
   */
  hubptr += (1 + obj->nPubs-1 + obj->nPris-1 + obj->nImps-1)<<2;
  
  /* imported objects will have their VAR blocks allocated after the importing
   * object's own VAR block variables); the DAT sections (if any) immediately
   * follow the header data, and the byte code for the methods follows the DAT
   * sections.
   */
  obj->vcurr = obj->vsize;
  obj->datptr = hubptr;
  hubptr += obj->datsize;
  hubptr = (hubptr + 3) & ~3;
}

static int wordbase;
static int bytebase;
static int firstPri;
static int firstImp;
static int firstDAT;

/* During pass 1 symbols are allocated from a zero base. At the end of pass 1
 * we can place the data one after the other in either in the PBASE or VBASE
 * segments; also PRI methods get an index sequentially following PUB indices
 * and imported objects get indices sequentially following PRI methods.
 */
static void rebase_symbols(Symbol* sym)
{
  if( !sym ) return;
  rebase_symbols(sym->left);
  rebase_symbols(sym->right);

  switch( sym->type ) {
    case DAT_SYMBOL:
      while(sym) {
        ((DATSym*)(sym+1))->hubloc += firstDAT;
        sym = ((DATSym*)(sym+1))->next_lcl;
      }
      break;
      
    case VAR_SYMBOL:
      if( sym->size==2 ) sym->value += wordbase;
      if( sym->size==1 ) sym->value += bytebase;
      break;
      
    case PRI_SYMBOL:
      sym->idx += firstPri-1;
      break;
      
    case OBJ_SYMBOL:
      sym->idx += firstImp-1;
      break;
  }
}

void EmitObjectTail(Object *obj)
{
  int tmp;
  ObjHdr *objhdr;
  
  if( pass==1 ) {
    /* For the VAR block, place word variables after long variables and byte
     * variables after that; round up the VAR block to a long alignment.
     * Calculate the header size (in bytes), which is the base for the DAT
     * section data. Adjust symbol addresses.
     */
    wordbase = obj->varOffsets[2];
    bytebase = wordbase + obj->varOffsets[1];
    obj->vsize = bytebase + obj->varOffsets[0];
    obj->vsize = (obj->vsize + 3) & ~3;
    firstPri = 1 + obj->nPubs-1;
    firstImp = firstPri + obj->nPris-1;
    firstDAT = (1 + obj->nPubs-1 + obj->nPris-1 + obj->nImps-1)<<2;
    if( firstDAT>1024 ) hsc_abort("Too many methods and imported objects");
    rebase_symbols(obj->Globals);

    /* In pass 1 vcurr counts the total space needed for the VAR blocks of
     * imported object instances; similarly, in pass 1 datptr counts the total
     * size of the DAT blocks.
     */    
    obj->impsize = obj->vcurr;
    obj->datsize = obj->datptr;
    obj->nImp2 = firstImp;
    return;
  }
  
  assert( obj->nImp2 == (1 + obj->nPubs-1 + obj->nPris-1 + obj->nImps-1) );
  assert( obj->vcurr  == (obj->vsize + obj->impsize) );
  tmp = (1 + obj->nPubs-1 + obj->nPris-1 + obj->nImps-1)<<2;
  assert( obj->datptr == (obj->pbase + obj->datsize + tmp) );
  
  /* align hubptr to a long boundary and fixup object length in the header */
  hubptr = (hubptr + 3) & ~3;
  objhdr = (ObjHdr*)(hubram + obj->pbase);
  objhdr->objlen = hubptr - obj->pbase;  
}

static int main_paddr;
static int main_stack;

void EmitMethodHdr(Symbol *method)
{
  PubLnk *link = (PubLnk*)(hubram + curobj->pbase);
  int idx = method->idx;
  
  if( pass==1 ) return;
  
  assert( method->type==PUB_SYMBOL || method->type==PRI_SYMBOL );
  link[idx].paddr  = hubptr - curobj->pbase;
  link[idx].lclsiz = method->size - method->argc*4 - 4;
  
  if( curobj==topobj && idx==1 ) {
    main_paddr = hubptr;
    main_stack = method->size;
  }
}

void EmitMethodTail(Symbol *method)
{
  /* nothing to do or check */
}

/* Allocate the VAR block for an imported object. These VAR blocks are placed
 * one after the other in the importing object's VBASE segment. The offset from
 * the start of the VBASE segment is entered in the import link. As the imported
 * object's PBASE may not be known yet, we leave an index in its place and fix
 * it up when all objects have been compiled (see code below).
 */
void AllocImportObj(Object *import)
{
  ObjLnk *objlnk;

  if( pass==1 ) {
    curobj->vcurr += import->vsize;
    return;
  }
  
  objlnk = (ObjLnk*)(hubram + curobj->pbase);
  objlnk[curobj->nImp2].idx = import->idx;
  objlnk[curobj->nImp2].vdiff = curobj->vcurr;
  curobj->vcurr += import->vsize;
  ++curobj->nImp2;
}

/* Fix up the object import links: walk the hubram datastructure to locate
 * each object import link and replace the object index by its relative
 * PBASE address.
 */
static Object* find_obj(int idx);

static void fixup_obj_plinks()
{
  int i, first, last;
  Object *obj, *target;
  ObjHdr *objhdr;
  ObjLnk *objlnk;
  
  for(obj = topobj; obj; obj = obj->next) {
    objhdr = (ObjHdr*)(hubram + obj->pbase);
    objlnk = (ObjLnk*)(hubram + obj->pbase);
    first = objhdr->obj1st;
    last  = first + objhdr->objcnt;
    for(i = first; i<last; ++i) {
      target = find_obj(objlnk[i].idx);
      objlnk[i].pdiff = target->pbase - obj->pbase;
    }
  }
}

static Object* find_obj(int idx)
{
  Object *obj;
  
  for(obj=topobj; obj; obj = obj->next) {
    if( obj->idx==idx ) break;
  }
  assert( obj ); /* internal error: object index not found */
  return obj;
}

void EmitProgramHdr()
{
  if( pass==1 ) {
    memset(hubram, 0, sizeof(hubram));
  }
  hubptr = sizeof(PgmHdr);
}

static int checksum();

#define HASFRQ 0
#define HASXIN 1
#define HASMOD 2

/* EmitProgramTail is called at the end of code generation and perhaps a
 * misnomer as most of the code is about generating the PNUT program header
 * structure.
 */
void EmitProgramTail()
{
  PgmHdr *hdr = (PgmHdr*)hubram;
  Symbol *sym;
  int i, clkfreq = 0, clkmode = 0, opts = 0;
  static char *con[3] = { "_CLKFREQ", "_XINFREQ", "_CLKMODE" };
  int val[3];

  if( pass==1 ) return;

  fixup_obj_plinks();

  /* read clock mode settings, if any */
  for(i=0; i<3; ++i) {
    sym = SymbolLookup(con[i], topobj->Globals);
    val[i] = 0;
    if( sym ) {
      if( sym->type!=INT_CON_SYMBOL) hsc_abort("clock setting not integer constants");
      val[i] = sym->value;
      opts |= 1<<i;
    }
  }
  if( (opts&HASFRQ && opts&HASXIN) || opts&(HASFRQ|HASXIN) && !opts&HASMOD) {
    hsc_abort("bad combination of clock settings");
  }
  
  /* the values of the _clkmode variable are different from those for clkmode
   * register, so we need to compute a mapping; also we need to work out the
   * clock frequency from _xinfreq, if _clkfreq is not given directly.
   * See pp. 69 and 93 of the Prop Manual v1.2 for details of the mapping.
   */
  if( val[2] & ~0x7ff ) hsc_abort("clock mode value out of range");
  switch( val[2] & 0x03f ) {
    case 0x001: /* rcfast */ clkmode = 0x00; clkfreq = 12000000; break;
    case 0x002: /* rcslow */ clkmode = 0x01; clkfreq =    20000; break;
    case 0x004: /* xinput */ clkmode = 0x20; break;
    case 0x008: /* xtall1 */ clkmode = 0x28; break;
    case 0x010: /* xtall2 */ clkmode = 0x30; break;
    case 0x020: /* xtall3 */ clkmode = 0x38; break;
    default: hsc_abort("illegal clock source combination");
  }
  if( clkfreq && (opts&HASFRQ) ) hsc_abort("cannot combine clock source RC with _CLKFREQ");
  if( val[2] & 0x7c0) {
    if( clkfreq ) hsc_abort("cannot combine clock source RC with PLL");
    clkfreq = val[1];
    switch( val[2] & 0x7c0 ) {
      case 0x040: /* pll1x  */ clkmode |= 0x43;                break;
      case 0x080: /* pll2x  */ clkmode |= 0x44; clkfreq *=  2; break;
      case 0x100: /* pll4x  */ clkmode |= 0x45; clkfreq *=  4; break;
      case 0x200: /* pll8x  */ clkmode |= 0x46; clkfreq *=  8; break;
      case 0x400: /* pll16x */ clkmode |= 0x47; clkfreq *= 16; break;
      default: hsc_abort("illegal PLL clock mode combination");
    }
  }
  else if( !clkfreq ) {
    clkmode |= 0x02;
  }
  if( opts&HASFRQ ) clkfreq = val[0];

  /* fill out the program header structure */  
  hdr->clkfreq = clkfreq;
  hdr->clkmode = clkmode;
  hdr->pbase = topobj->pbase;
  hdr->vbase = hubptr;
  hdr->dbase = hubptr + topobj->vsize + topobj->impsize + 8;
  hdr->pcurr = main_paddr;
  hdr->dcurr = hdr->dbase + main_stack;
  hdr->checksum = -checksum();
}

static int checksum()
{
  unsigned char sum = 0xec;
  int i;
  
  for(i=0; i<hubptr; i++) sum = sum + hubram[i];
  return sum;
}

void SaveBinary(char *path)
{
  PgmHdr *hdr = (PgmHdr*)hubram;
  FILE *f;
  
  f = fopen(path, "wb");
  if( !f ) hsc_abort("could not open output file");
  fwrite(hubram, hdr->vbase, 1, f);
  fclose(f);
}
