
/* Hive Spin Compiler OBJ block parsing */
/* This code is public domain */

#include "keyword.h"
#include "symbol.h"
#include "parse.h"
#include "hsc.h"
#include <stdlib.h>
#include <string.h>
#include <assert.h>

#define MAXPATH 4096
static char path[MAXPATH];

/* Objects are linked in a chain: 'topobj' is the main object that started
 * compilation, 'curobj' is the current object and 'lastobj' points to the
 * last object in the chain (or is null).
 */
Object *topobj;
Object *curobj;
Object *lastobj;

/* Retrieve the Object structure for object 'path'. There is only one
 * singleton object structure for each unique path.
 */
Object* NewObject(char* path)
{
  Object *obj;
  
  /* first look if this object has been loaded already */
  for(obj = topobj; obj; obj = obj->next) {
    if( !strcmp(obj->path, path) ) return obj;
  }
  
  /* allocate a new object structure, and link it to the object chain */
  obj = (Object*)malloc(sizeof(Object));
  if( !obj ) hsc_abort("out of memory");
  memset(obj, 0, sizeof(Object));
  strncpy(obj->path, path, MAXPATH);
  obj->nPubs = obj->nPris = obj->nImps = 1;
  obj->Globals = SymInit();
  obj->idx = 1;
  if( lastobj ) {
    obj->idx = lastobj->idx + 1;
    lastobj->next = obj;
  }
  lastobj = obj;
  return obj;
}

void ParseOBJpass1()
{
  int count, i;
  Symbol *sym;
  Object *obj;
  
  MustSkip(TK_OBJ);
  TestAndSkip(TK_EOL);
  
  while( IsId(token.type) ) {
    /* create a new symbol for the imported object in the global namespace
     * of the current object.
     */
    sym = AddSymbol(token.text, &(curobj->Globals));
    sym->type = OBJ_SYMBOL;
    NextToken();
    
    /* get the object instance count; default to 1 */
    count = 1;
    if( TestAndSkip(TK_LBRACKET) ) {
      count = EvalExpr(NULL);
      if( count<0 || count>255 ) hsc_abort("Too many imported objects");
      MustSkip(TK_RBRACKET);
    }
    MustSkip(TK_COLON);
    
    /* read the object name/path into the 'path' string */
    i = 0;
    do {
      if( !IsIntLiteral(token.type) ) hsc_abort("syntax error");
      path[i++] = token.value;
      NextToken();
    } while( TestAndSkip(TK_COMMA) );
    path[i] = 0;
    if( !strrchr(path, '.') ) strcat(path, ".spin");
    MustSkip(TK_EOL);
    
    /* now create a reference to the imported object; note that there is
     * only one 'Object' structure for each unique path -- information specific
     * to this import is stored in the symbol structures.
     */
    obj = NewObject(path);
    sym->obj = obj;
    sym->idx = curobj->nImps;
    sym->size = count;
    curobj->nImps += count;
    
    /* if this object has not been parsed for pass 1 data before, recursively
     * call the parser to collect it; after parsing the imported object,
     * restore the compiler object state to the current object.
     */
    if( !obj->pass1 ) {
      Object *savobj = curobj;
      void *ls = GetLexState();
      curobj = obj;
      ParsePass1(path);
      obj->pass1 = 1;
      PutLexState(ls);
      curobj = savobj;
      for(i=0; i<count; ++i) {
        AllocImportObj( obj );
      }
    }
  }
}

/*
 * Parsing an OBJ block during pass 2 is mostly eating through the syntax;
 * However, for each imported object there is a call into the PNUT code
 * generator to allocate space in the VBASE segment and to update the object
 * link table accordingly.
 * 
 * Note that a second pass over the imported object's source will
 * be made from the top level parsing routine, not through a recursive call
 * as in pass 1.
 */
void ParseOBJpass2()
{
  int i, count;
  Symbol *sym;
  
  MustSkip(TK_OBJ);
  TestAndSkip(TK_EOL);
  while( IsId(token.type) ) {
    sym = SymbolLookup(token.text, curobj->Globals);
    assert( sym );
    NextToken();
    
    count = 1;
    if( TestAndSkip(TK_LBRACKET) ) {
      count = EvalExpr(NULL);
      MustSkip(TK_RBRACKET);
    }
    assert( count==sym->size );
    MustSkip(TK_COLON);
    
    do {
      NextToken();
    } while( TestAndSkip(TK_COMMA) );
    MustSkip(TK_EOL);
    
    /* call AllocImportObj for each entry of an object array separately,
     * as each instance has its own VBASE data and its own link table
     * entry.
     */
    for(i=0; i<count; ++i) {
      AllocImportObj( sym->obj );
    }
  }  
} 
