/* Hive Spin Compiler Object and AST structure headers */
/* This code is public domain */

void ParsePass1(char *path);
void ParsePass2(char *path);

struct _object {

  /* parser */
  Symbol  *Globals;
  int      nPubs;
  int      nPris;
  int      nImps;
  int      varOffsets[3];
  int      datptr;
  int      pass1;
  
  /* backend */
  int      datsize;
  int      vsize;
  int      impsize;
  int      nImp2;
  int      pbase;
  int      pcurr;
  int      vcurr;
  
  /* object chain */
  int      idx;
  struct _object *next;
  char     path[4096];
};
typedef struct _object Object;

extern Object *topobj;  /* the object that started compilation, first in the list */
extern Object *curobj;  /* the current object */
extern Object *lastobj; /* the last object in the linked list */


/* The compiler first parses the tokens of a PUB or PRI section
 * into an abstract syntax tree, AST.
 * http://en.wikipedia.org/wiki/Abstract_syntax_tree
 * The tree has three types of nodes: Node, CaseItem and
 * RepeatBlock. Expressions are held entirely in Nodes. CaseItem
 * and RepeatBlock are convenience structures for representing
 * CASE and REPEAT statements respectively.
 */

struct _node {
  int op;
  union {                    /* node type: TK_...             */
    struct _node *sub;       /* sub node                      */
    Symbol *sym;             /* object or method symbo        */
    int val;                 /* value for type TK_INTLITERAL  */
  };
  struct _node *nxt;         /* next node                     */
};
typedef struct _node Node;

Node* new_node(int op, Node *sub);
Node* new_node2(int op, Node *sub, Node *nxt);

struct _caseitem {
  int op;                    /* TK_CASE ore TK_OTHER           */
  Node *le1;                 /* <expr1>:                       */
  Node *le2;                 /* <expr1>..<expr2>:              */
  Node *body;                /* statements for this case       */
  int   lbl;                 /* code generator label for block */
  struct _caseitem *next;    /* case items form a linked list  */
};
typedef struct _caseitem CaseItem;

struct _repeatblock {
  int op;                     /* TK_RPT, TK_RPTWHILE, etc.     */
  Node *e;                    /* condition expression          */
  Node *from;                 /* from expression               */
  Node *to;                   /* to expression                 */
  Node *step;                 /* step expression               */
  Node *body;                 /* repeat body                   */
  int   next;                 /* codegen label for next        */
  int   quit;                 /* codegen label for quit        */
  int   caselevel;            /* codegen caselevel counter     */
  struct _repeatblock *enclosing;  /* enclosing rpt, if any    */
};
typedef struct _repeatblock RepeatBlock;

Node* parse_arglist(int argc);
Node* ParseVoidExpr();
Node* ParseExpr();
void EmitByte(int c);
void EmitBlock(Node* n);

/* accesss types */
#define VOID          0x04
#define PUSH          0x00
#define STORE         0x01
#define LEA           0x02
#define PEA           0x03

#define EA_BIT   0
#define EA_BYTE  2
#define EA_WORD  4
#define EA_LONG  6

#define WITH_PUSH     0x80

/* The PNUT interpreter accesses hubram relative to one of four
 * base values:
 * - relative to 0, i.e. absolute addresses; this is used for
 *   accesses like LONG[<expr>]
 * - relative to pbase, i.e. the start of the current object; this
 *   is used for jump addresses in code and for DAT block access
 * - relative to vbase, i.e. the start of the current VAR block
 *   (note that each object instance has its own VAR block)
 * - relative to dbase, i.e. the start of the current stack frame
 *   with parameters and local variables
 */
#define MEM           0x00  // base is 0, offset is on stack
#define DAT           0x04  // base is pbase, offset follows opcode
#define VAR           0x08  // base is vbase, offset follows opcode
#define LCL           0x0c  // base is dbase, offset follows opcode
 
/* The base values are kept in 3 interpreter registers. Two additional
 * registers are used to keep the current code location and the
 * current stack pointer:
 * 0xb - PBASE - start of current object       (~ 'gp')
 * 0xc - VBASE - start of current VAR block    (~ 'this')
 * 0xd - DBASE - start of current stack frame  (= 'bp')
 * 0xe - PCURR - current code location         (= 'pc')
 * 0xf - DCURR - current top of stack          (= 'sp')
 * 
 * Object code at PBASE is long aligned and has the following
 * structure:
 * - a long with a pointer to the next object.
 * - a long for each PUB/PRI method; the upper word contains the
 *   address of the method relative to PBASE and the lower word
 *   contains the size of the local variables (count * 4).
 * - a long for each imported object; the upper word contains the
 *   offset from the current PBASE to the imported object's PBASE
 *   and the lower word contains the offset from the current VBASE
 *   to the imported object's VBASE (note that the compiler can
 *   calculate this because objects are statically instantiated).
 * - object arrays have one long as per above for each element
 * - the compiled DAT block
 * - the byte code for the PUB/PRI methods of the object
 * - padding to the next long as required
 */
 
