
/* Hive Spin Compiler PNUT code generator & assembler */
/* This code is public domain */

#include "symbol.h"
#include "keyword.h"
#include "hsc.h"
#include "parse.h"
#include <stdio.h>
#include <assert.h>

int pnutpass;

void emit_byte(int c)
{
  static int ctr = 16;
  
  if( pnutpass==2 ) {
    c &= 0xff;
    printf(" %02x", c);
    if( !--ctr ) {
      ctr = 16;
      printf("\n");
    }
  }
  EmitByte( c );
}

static int get_pcurr()
{
  return hubptr - curobj->pbase;
}

/* PNUT has variable sized instructions to push an address on
 * the stack or to jump to an address, so called 'span dependent
 * instructions' (SDIs for short). Assembling such programs to
 * a minimal size in the general case is an NP-complete problem.
 * When there is no label arithmetic or code rearrangement the
 * following algorithm runs in O(n^3) in a degenerate case, but
 * in O(n^2) for real world programs (Szymanksi 1978):
 * 
 * Run the first pass assuming that each SDI is long, noting the
 * location and target of each SDI in a table. Then between the first
 * and second passes, iterate over the table looking for an instruction
 * that is long but could be short, shorten it, and adjust symbol table
 * values appropriately. Repeat until you don't find any more.
 * 
 * During pass 1 forward labels are represented by a negative address.
 * The resolve routine scans the table for that negative number and
 * updates the target address to its intended value.
 */
 
#define LBL  1

struct _sdi_labels {
  int loc;       /* location of SDI */
  int tgt;       /* target of SDI   */
  char typ;      /* JMP or LBL      */
  char siz;      /* 2 or 3 byte op  */
};
typedef struct _sdi_labels Label;

/* For now use a fixed length table FIXME */
static Label labels[1000];
static int lblno;

static int fwdlabel;
#define NEWFWD --fwdlabel

/* 'done' is shared between the next two routines and signals completion
 * of the Szymanski algorithm.
 */
static int done;

void adjust_SDI_tab(int idx)
{
  int i, loc;
  Label *label;
  
  labels[idx].siz = 2;
  loc = labels[idx].loc;
  for(i=0; i<lblno; ++i) {
    label = labels + i;
    if( label->loc > loc ) --label->loc;
    if( label->tgt > loc ) --label->tgt;
  }
  done = 0;
}

/* PNUT SDIs are either 2 byte or 3 byte. Note that the 'PUSH <mask>'
 * instruction cannot be used for addresses, because it violates the
 * constraints of the Szymanksi algorithm.
 */
void resize_SDIs()
{
  int i, len;
  Label *label;
  
  do {
    done = 1;
    for(i=0; i<lblno; ++i) {
      label = labels + i;
      if( label->siz==2 ) continue;
      if( label->typ==LBL ) {
        if( label->tgt>0 && label->tgt<255 ) adjust_SDI_tab(i);
      }
      else { /* jmp, jnz, caseval, etc. */
        len = label->tgt - (label->loc+2);
        if( len>-64 && len<64 ) adjust_SDI_tab(i);
      }
    }
  } while( !done );
}

static int emit_addr(int tgt)
{
  Label *lbl = labels + lblno++;
  
  if( pnutpass==1 ) {
    lbl->typ = LBL;
    lbl->siz = 3;
    lbl->loc = get_pcurr();
    lbl->tgt = tgt; 
    emit_byte(0); emit_byte(0); emit_byte(0);
    return tgt;
  }
  if( lbl->loc!=get_pcurr() ) hsc_abort("internal error, pnutpass 2 label mismatch");
  if( lbl->siz==2 ) {
    emit_byte(0x38); emit_byte(lbl->tgt);
    return lbl->tgt;
  }
  emit_byte(0x39); emit_byte(lbl->tgt>>8); emit_byte(lbl->tgt&0xff);
  return lbl->tgt;  
}

static int emit_jump(int op, int tgt)
{
  Label *lbl = labels + lblno++;
  int len;
  
  if( pnutpass==1 ) {
    lbl->typ = op;
    lbl->siz = 3;
    lbl->loc = get_pcurr();
    lbl->tgt = tgt; 
    emit_byte(0); emit_byte(0); emit_byte(0);
    return tgt;
  }
  if( lbl->loc!=get_pcurr() ) hsc_abort("internal error, pnutpass 2 label mismatch");
  if( tgt>=0 && lbl->tgt!=tgt ) hsc_abort("internal error, pnutpass 2 label mismatch");
  emit_byte(op);
  len = lbl->tgt - (lbl->loc+2);
  if( lbl->siz==2 ) {
    if( len<0 ) len += 0x80;
    emit_byte(len);
    return lbl->tgt;
  }
  len -= 1;
  emit_byte( (len>>8) | 0x80 );
  emit_byte( len & 0xff );
  return lbl->tgt;
}

static void resolve(int label)
{
  int i;
  
  if( pnutpass==1 ) {
    assert( label<0 );
    for(i=0; i<lblno; ++i) {
      if( labels[i].tgt==label ) labels[i].tgt = get_pcurr();
    }
  }
}

static void emit_stmt(Node *n);
static void emit_exit(Node* n);
static void emit_if(Node *n);
static void emit_case(Node *n);
static void emit_qnxt(Node *n);
static void emit_repeat(Node *n);

static void emit_expr(Node* n);
static int  emit_expr2(Node* n, int mode);
static void emit_push_int(int val);

/* EmitBlock is the main entry into the code generator. It makes
 * two passes over the AST. In the first pass the SDI/label table
 * is built up, and in the second pass the code is generated.
 */
void EmitBlock(Node *n)
{
  int sav = hubptr;
  
  pnutpass = 1;
  lblno = 0;
  fwdlabel = 0;
  if( n ) emit_stmt(n);

  resize_SDIs();
  
  hubptr = sav;
  pnutpass = 2;
  lblno = 0;
  if( n ) emit_stmt(n);
  emit_byte(0x32);
}

static void emit_stmt(Node *n)
{
  switch( n->op ) {
    case TK_ABORT:
    case TK_RETURN:    emit_exit(n);    break;
    case TK_IFNOT:
    case TK_IF:        emit_if(n);      break;
    case TK_CASE:      emit_case(n);    break;
    case TK_RPT:
    case TK_WHILE:
    case TK_UNTIL:
    case TK_RPTWHILE:
    case TK_RPTUNTIL:
    case TK_RPTCNT:
    case TK_RPTFROM:   emit_repeat(n);  break;
    case TK_NEXT:
    case TK_QUIT:      emit_qnxt(n);    break;
    default:           emit_expr(n);    break;
    case TK_LIST:
      while( n ) {
        emit_stmt(n->sub);
        n = n->nxt;
      }
      break;
  }
}

static void emit_exit(Node* n)
{
  if( n->sub ) {
    emit_expr(n->sub);
    emit_byte( (n->op==TK_RETURN) ? 0x33 : 0x31 );
  }
  else {
    emit_byte( (n->op==TK_RETURN) ? 0x32 : 0x30 );
  }
}

#define JMP     0x04
#define JZ      0x0a
#define JNZ     0x0b
#define TJZ     0x08
#define DJNZ    0x09
#define RPTINC  0x02
#define RPTSTEP 0x06

static void emit_if(Node *n)
{
  int op, jmp1, jmp2;

  emit_expr(n->sub);
  op = (n->op==TK_IFNOT) ? JNZ : JZ;
  jmp1 = emit_jump(op, NEWFWD);
  assert( n->nxt );
  if( n->nxt->op==TK_ELSE ) {
    emit_stmt(n->nxt->sub);
    jmp2 = emit_jump(JMP, NEWFWD);
    resolve(jmp1);
    emit_stmt(n->nxt->nxt);
    resolve(jmp2);
  }
  else {
    emit_stmt(n->nxt);
    resolve(jmp1);
  }
}

/* Each active case has two words on the stack. If we EXIT or QUIT out of
 * a CASE statement enclosed in a REPEAT, these need to be removed. With
 * caselevel we track the number of nested CASE statements inside of the
 * innermost enclosing REPEAT statement.
 */
static int caselevel = 0;

/* A case statement first pushes its exit ('break') PBASE address and
 * the case expression. These remain on stack throughout. Next for each
 * case item, the value or range is pushed and tested against the 
 * expression (popping the value or range). If a match occurs, the test
 * jumps to the corresponding body, else it falls through. Each path
 * ends with a CASEEND opcode which pops the case expression and the
 * exit address, jumping to the latter.
 */
static void emit_case(Node *n)
{
  int end, op;
  CaseItem *ci;
  
  ++caselevel;
  
  /* first push the exit addr and the case expression */
  end = emit_addr(NEWFWD);
  emit_expr( n->sub );

  /* next generate the case expressions */
  for(ci=(CaseItem*)(n->nxt); ci; ci = ci->next) {
    if( !ci->le1 ) break; /* = case OTHER */
    op = 0x0d; /* CASEVAL */
    emit_expr(ci->le1);
    if( ci->le2 ) {
      emit_expr(ci->le2);
      op = 0x0e; /* CASERNG */
    }
    ci->lbl = emit_jump(op, NEWFWD);
  }
  
  /* then generate the other case body */
  if( ci ) emit_stmt(ci->body);
  emit_byte(0x0c); /* CASEEND */
  
  /* end with generating the normal bodies */
  for(ci=(CaseItem*)(n->nxt); ci; ci = ci->next) {
    if( !ci->le1 ) break;
    resolve(ci->lbl);
    if( ci->body ) emit_stmt(ci->body);
    emit_byte(0x0c); /* CASEEND */
  }
  resolve(end);
  
  --caselevel;
}

static RepeatBlock *current_rpt = NULL;

/* Handle 'exit' and 'quit'; the parser has already checked that 
 * there is an enclosing repeat. First check for case stack variables
 * that must be popped if CASE statements are exited. The see if a counted
 * repeat is quit, because then the counter must be popped. This is done
 * using JNZ (note that the counter will never be zero).
 */
static void emit_qnxt(Node *n)
{
  int op;
  if( caselevel ) {
    emit_push_int( caselevel<<3 );
    emit_byte(0x14); /* POP   n */
  }
  if( n->op==TK_QUIT ) {
    op = (current_rpt->op==TK_RPTCNT) ? JNZ : JMP;
    emit_jump(op, current_rpt->quit);
    return;
  }
  emit_jump(JMP, current_rpt->next);
}

static void emit_repeat(Node *n)
{
  int op, loop;
  RepeatBlock *rpt = current_rpt = (RepeatBlock*)n;

  rpt->caselevel = caselevel;
  caselevel = 0;

  rpt->next = NEWFWD;
  rpt->quit = NEWFWD;
  switch( rpt->op ) {
    /* case 1 is simply a jump back to the top of the loop */
    case TK_RPT:
      rpt->next = get_pcurr();
      if( rpt->body) emit_stmt( rpt->body );
      emit_jump(JMP, rpt->next);
      break;
    /* case 2 and 3 are conditional jumps back to the top */
    case TK_RPTWHILE:
    case TK_RPTUNTIL:
      loop = get_pcurr();
      if( rpt->body) emit_stmt( rpt->body );
      resolve(rpt->next);
      emit_expr( rpt->e );
      op = (rpt->op==TK_RPTWHILE) ? JNZ : JZ;
      emit_jump(op, loop);
      break;
    /* case 4 and 5 are an exit test at the top, followed by case 1 */
    case TK_WHILE:
    case TK_UNTIL:
      rpt->next = get_pcurr();
      emit_expr( rpt->e );
      op = (rpt->op==TK_WHILE) ? JZ : JNZ;
      rpt->quit = emit_jump(op, NEWFWD);
      if( rpt->body) emit_stmt( rpt->body );
      emit_jump(JMP, rpt->next);
      break;
    /* case 6: push the count and skip body if already zero;
     * otherwise, do the body and jump back using 'decrement
     * and jump if not zero'. Note that TJZ and DJNZ only pop the
     * counter value when zero is reached; emit_qnxt deals with
     * popping the counter upon QUIT.
     */
    case TK_RPTCNT:
      emit_expr( rpt->e );
      emit_jump(TJZ, rpt->quit);
      loop = get_pcurr();
      if( rpt->body) emit_stmt( rpt->body );
      resolve(rpt->next);
      emit_jump(DJNZ, loop);
      break;
    /* case 7 is complicated. First we assign 'start' to the counter
     * variable and emit the body. Each iteration requires 'from' and
     * 'to' on the stack, and possibly 'step'. Two extended operations
     * (RPTADD and RPTINC) adjust the counter variable and jump back
     * to the start of the loop as long as it is in range.
     */ 
    case TK_RPTFROM:
      emit_expr( rpt->from );
      emit_expr2( rpt->e, STORE);
      loop = get_pcurr();
      if( rpt->body) emit_stmt( rpt->body );
      resolve( rpt->next );
      if( rpt->step ) emit_expr(rpt->step);
      emit_expr(rpt->from);
      emit_expr(rpt->to);
      op = (rpt->step) ?  RPTSTEP : RPTINC;
      emit_expr2(rpt->e, LEA);
      emit_jump(op, loop);
      break;
  }
  resolve(rpt->quit);
  caselevel = rpt->caselevel;
  current_rpt = rpt->enclosing;  
}

static void emit_mem_op(int size, Node *index, int base, int access);
static int  emit_var_op(Node* n, int access);
static void emit_builtin(Node *n, int mode);
static void emit_lookup(Node *n);
static void emit_call(Node *call, Node *obj);
static void setup_pnut(Node *n);

static void emit_args(Node *n)
{
  while(n) {
    emit_expr(n->sub);
    n = n->nxt;
  }  
}

void emit_expr(Node* n)
{
  emit_expr2(n, PUSH);
}

/* The expression code generator operates in one of 5 modes:
 * 4 VOID  - do not leave a value on stack
 * 0 PUSH  - calculate/fetch a value, leave result on stack (= default mode)
 * 1 STORE - locate lval, pop value and store in lval
 * 2 LEA   - locate lval, shift PNUT into EA mode
 * 3 PEA   - calculate lval addr, push result on stack
 * See pnut.h for more detail.
 */
int emit_expr2(Node* n, int mode)
{
  int size, op;
  
  /* first we handle lval nodes, i.e. anything that can hold a value */
  switch( n->op ) {
    case DAT_SYMBOL:
    case VAR_SYMBOL:
    case LCL_SYMBOL:
      size = emit_var_op(n, mode);
      return size;

    case TK_BYTE:
    case TK_WORD:
    case TK_LONG:
      size = 1 << (n->op - TK_BYTE);
      emit_expr( n->sub );
      emit_mem_op(size, n->nxt, MEM, mode);
      return size;

    case TK_SPR:
      assert( mode!=PEA );
      emit_expr(n->sub);
      emit_byte(0x24 + mode);
      return 4;
    
    case  TK_PAR:  case  TK_CNT:  case  TK_INA:  case  TK_INB: 
    case  TK_OUTA: case  TK_OUTB: case  TK_DIRA: case  TK_DIRB:
    case  TK_CTRA: case  TK_CTRB: case  TK_FRQA: case  TK_FRQB:
    case  TK_PHSA: case  TK_PHSB: case  TK_VCFG: case  TK_VSCL:
      assert( mode!=PEA );
      size = 4;
      op = 0x3f;
      if( n->sub ) {
        size = 0;
        op = 0x3d;
        emit_expr( n->sub );
      }
      if( n->nxt ) { 
        op = 0x3e;
        emit_expr( n->nxt );
      }
      emit_byte(op);
      emit_byte( 0x90 | GetReg(n->op) | mode<<5 );
      return size;
  }

  /* for all other node types STORE, LEA and PEA are illegal modes */
  assert( mode==VOID || mode==PUSH );    
  
  /* handle the builtin constants and functions */
  if( IsBuiltin(n->op) ) {
    emit_builtin(n, mode);
    return 0;
  }

  /* next handle the special cases */
  switch( n->op ) {
    
    case TK_INTLITERAL:
    case TK_FLOATLITERAL:
      assert(mode==PUSH);
      emit_push_int( n->val );
      return 0;

    case TK_ATAT:
      /* Calculate as address of DAT block byte array */
      assert(mode==PUSH);
      emit_expr(n->sub);
      emit_byte(0x97); emit_byte(0x00);  /* FETCH.B   @DAT[0][*] */
      return 0;

    case TK_AT:
      /* Calculate as address of DAT block byte array */
      assert(mode==PUSH);
      emit_expr2(n->sub, PEA);
      return 0;

    case TK_VOID:
      /* switch from PUSH to VOID mode */
      assert(mode==PUSH);
      emit_expr2(n->sub, VOID);
      return 0;
    
    case TK_COLONEQUAL:
      /* keeping the value on stack following storing it is only
       * available as an EA operation.
       */
      emit_expr(n->nxt);
      if( mode==VOID ) return emit_expr2(n->sub, STORE);
      emit_expr2(n->sub, LEA);
      emit_byte(0x80); /* EA_STORE WP */
      return 0;

    case OBJ_SYMBOL:
      emit_call(n->sub, n);
      return 0;
      
    case TK_FRAME0:    case TK_FRAME1:
    case TK_FRAME2:    case TK_FRAME3:
      emit_call(n, NULL);
      return 0;

    case TK_LIST:
      /* push actuals list to a method or builtin function call */
      for(; n; n = n->nxt ) emit_expr(n->sub);
      return 0;
  }
  
  /* Compound assignments have two steps: first we push the argument
   * and load the effective address of the lval; next we emit the
   * EA variant of the binary operator. See pnut.h for details.
   */
  if( IsAssignmentOp(n->op) ) {
    emit_expr(n->nxt);
    emit_expr2(n->sub, LEA);
    op = GetSpinOpcode(n->op) & 0x5f;
    if( mode==PUSH ) op |=0x80;
    emit_byte( op );
    return 0;
  }
  
  /* If we reach this point we must be dealing with an operator.
   * Binary operators are simple: push the arguments and emit the
   * operator opcode.
   */
  if( IsBinaryOp(n->op) ) {
    assert(mode==PUSH);
    emit_expr(n->sub);
    emit_expr(n->nxt);
    emit_byte( GetSpinOpcode(n->op) );
    return 0;
  }
  
  /* For pre-/suffix ops we distinguish between pre and post variants. For
   * a pre variant the lval is in 'sub', for post in 'nxt'. The PRE/POST
   * INC/DEC operators require a size indication of the lval and this is
   * or'ed into the EA operator.
   */
  if( IsPostfixOp(n->op) ) {
    if( n->sub ) {
      size = emit_expr2(n->sub, LEA);
      op = GetSpinOpcode(n->op);
    }
    else {
      size = emit_expr2(n->nxt, LEA);
      op = GetSpinOpcode2(n->op);
    }
    if( n->op==TK_PLUSPLUS || n->op==TK_MINUSMINUS ) op |= (size==4 ? 6 : size<<2);
    if( mode==PUSH ) op |= 0x80;
    emit_byte(op);
    return 0;
  }

  if( n->op==TK_UMINUS && n->sub->op==TK_INTLITERAL ) {
    emit_push_int( -(n->sub->val) );
    return 0;
  }
  
  if( IsUnaryOp(n->op) ) {
    op = GetSpinOpcode(n->op);
    if( mode==VOID ) {
      emit_expr2(n->sub, LEA);
      emit_byte(op & 0x5f);
      return 0;
    }
    emit_expr2(n->sub, PUSH);
    emit_byte(op);
    return 0;
  }
  
  hsc_abort("internal error: unhandled node in emit_expr");
  return 0;
}


void emit_mem_op(int size, Node *index, int area, int mode)
{
  int op;
  
  op  = 0x80 | ((size>>1)<<5) | area | mode;
  if( index ) {
    emit_expr( index );
    op |= 0x10;
  }
  emit_byte( op );
}

int emit_var_op(Node* n, int mode)
{
  static char membase[] = { DAT, VAR, LCL };
  Symbol *sym = (Symbol*)(n->sub);
  int op, offset, area = membase[GetVarArea(n->op)];
  
  /* see if a short form can be used */
  if( area!=DAT && sym->size==4 && sym->offset<32 && !n->nxt ) {
    op = (area==VAR) ? 0x40 : 0x60;
    op |= sym->offset | mode;
    emit_byte( op );
    return 4;  
  }
  
  /* long form consists of the general memory access opcode, followed
   * by one or two (MSB first) offset bytes; note that a potential
   * extended operation code comes after the offset bytes.
   */
  emit_mem_op(sym->size, n->nxt, area, mode );
  offset = (area==DAT) ? ((DATSym*)(sym+1))->hubloc : sym->offset;
  if( offset<0x80 ) {
    emit_byte( offset );
  }
  else {
    emit_byte( offset>>8 | 0x80 );
    emit_byte( offset & 0xff );
  }
  return sym->size;
}

void emit_builtin(Node *n, int mode)
{
  Node* param2;
  int op;
  
  switch( n->op ) {
    case TK_COGINIT:
      assert( n->sub && n->sub->nxt );
      param2 = n->sub->nxt->sub;
      if( param2->op!=PUB_SYMBOL && param2->op!=PRI_SYMBOL ) {
        emit_args( n->sub );
      }
      else {
        setup_pnut( n );
      }
      emit_byte( (mode==PUSH) ? 0x28 : 0x2c );
      break;

    case TK_LOCKNEW:
    case TK_LOCKSET:
    case TK_LOCKCLR:
      if( n->sub ) emit_args( n->sub );
      op = GetSpinOpcode(n->op);
      op |= (mode==VOID) ? 4 : 0;
      emit_byte(op);
      break;

    case TK_CHIPVER:
      /* The chip version is stored at byte 0xffff in hubrom */
      emit_byte(0x34); /* PUSH       #-1    */
      emit_byte(0x80); /* FETCH.B    MEM[*] */
      break;
      
    case TK_CLKFREQ:
      /* The freqency is stored at long 0x0000 in hubram */
      emit_byte(0x35); /* PUSH       0      */
      emit_byte(0xc0); /* FETCH.L    MEM[*] */
      break;
      
    case TK_CLKMODE:
      /* The clock mode is stored at byte 0x0004 in hubram */
      emit_push_int(4); /* PUSH      #4     */
      emit_byte(0x80);   /* FETCH.B   MEM[*] */
      break;
      
    case TK_COGID:
      /* The cogid is in PNUT register 0x9 */
      emit_byte(0x3f); emit_byte(0x89);  /* FETCH.L   REG[ID] */
      break;

    case TK_STRCOMP:
    case TK_STRSIZE:
      emit_args(n->sub);
      emit_byte( GetSpinOpcode(n->op) );
      break;

    case TK_LOOKUP:
    case TK_LOOKUPZ:
    case TK_LOOKDOWN:
    case TK_LOOKDOWNZ:
      emit_lookup(n);
      break;
    
    /* builtin functions that are always void */    
    case TK_BYTEFILL:    case TK_WORDFILL:    case TK_LONGFILL:
    case TK_WAITPEQ:     case TK_BYTEMOVE:    case TK_WORDMOVE:
    case TK_LONGMOVE:    case TK_WAITPNE:     case TK_CLKSET:
    case TK_COGSTOP:     case TK_LOCKRET:     case TK_WAITCNT:
    case TK_WAITVID:
      emit_args(n->sub);
      emit_byte( GetSpinOpcode(n->op) );
      break;

    case TK_REBOOT:
      /* REBOOT is the same as CLKSET($80, 0) */
      emit_push_int(4); /* PUSH      #$80   */
      emit_push_int(4); /* PUSH      #0     */
      emit_byte(0x20);   /* CLKSET    *,*    */
      break;
  }
}

/* Check if the value fits one of the mask values that PNUT encodes in one
 * byte according to the following scheme:
 * value            mask byte  
 *   2^(b+1)    =>  000bbbbb
 *   2^(b+1)-1  =>  001bbbbb
 * ~(2^(b+1))   =>  010bbbbb
 * -(2^(b+1))   =>  011bbbbb
 * The algorithm simply tests all 124 possibilities. If there is a match,
 * the mask or'ed with 0x80 is returned, or 0 when there is no match.
 */
static int test_mask(int val)
{
  int e = 2, mask;
  /* we don't test for mask==31 as those are effectively tests for 0 and -1 */
  for(mask=0; mask<31; ++mask) {
    if( val==e )     { mask |= 0x80; break; }
    if( val==(e-1) ) { mask |= 0xa0; break; }
    if( val==~e )    { mask |= 0xc0; break; }
    if( val==-e )    { mask |= 0xe0; break; }
    e <<= 1;
  }
  return( (mask & 0x80) ? mask : 0 );
}

/* Emit code to push a single int value onto the stack, ie.
 * "PUSH  #<val>". On PNUT this can take the following forms:
 * 0x34  PUSH.T                                   // push #-1
 * 0x35  PUSH.Z                                   // push #0
 * 0x36  PUSH.O                                   // push #1
 * 0x37  PUSH.M   <mask>                          // see test_mask() below
 * 0x38  PUSH.1   <byte>                          // push 1 byte value
 * 0x39  PUSH.2   <byte><byte>                    // push 2 byte value
 * 0x3a  PUSH.3   <byte><byte><byte>              // push 3 byte value
 * 0x3b  PUSH.4   <byte><byte><byte><byte>        // push 4 byte value
 * In each case a long is pushed, as the stack only holds longs.
 * The PUSH.1-3 instructions do not sign extend; to push a negative
 * value the two's complement is pushed followed by a bitwise not op.
 */
static void emit_push_int(int val)
{
  int msk;
  
  if( val==-1 ) {
    emit_byte(0x34);
  }
  else if( val==0 ) {
    emit_byte(0x35);
  }
  else if( val==1 ) {
    emit_byte(0x36);
  }
  else if( /*(val<0 || val>128) && fixme */ (msk=test_mask(val)) ) {
    emit_byte(0x37);
    emit_byte( msk & 0x7f );
  }
  else {
    /* figure out how many bytes to send */
    int i = 4, minus = 0;
    if( val<0 ) {
      minus = 1;
      val = ~val;
    }
    if( val<0xff )          i = 1;
    else if( val<0xffff )   i = 2;
    else if( val<0xffffff ) i = 3;
    else if( minus ) {
      minus = 0;
      val = ~val;
    }
    /* emit in big-endian order */
    emit_byte(0x37+i);
    val <<= (4-i)*8;
    for(; i; --i) {
      emit_byte( val>>24 );
      val <<= 8;
    }
    if( minus ) emit_byte(0xe7);
  }
}

/* A call occurs in three steps: first a FRAMEx op is emitted, which
 * pushes four words with PNUT registers (PBASE, VBASE, DBASE and a
 * placeholder word) and a zero intialised long (the callee return value).
 * PBASE is long aligned, and the bottom two bits encode whether the
 * caller expects a return value and whether the caller catches exceptions.
 * 
 * Next the actual parameters are calculated and pushed on to the stack
 * left to right (i.e. last parameter is on top).
 * 
 * Then one of three CALL ops is emitted. This fills the placeholder with
 * the return PCURR address and jumps via the index table at the start
 * of the object to the callee routine, adjusting PBASE and VBASE as needed.
 */
static void emit_call(Node *call, Node *obj)
{
  int oidx, idx = call->sym->idx;
  
  /* emit FRAME* op, push arguments */
  emit_byte( call->op & 3 );
  if( call->nxt ) emit_args( call->nxt );
  
  if( !obj ) {    
    /* emit CALL idx */
    emit_byte(0x05); emit_byte(idx);
    return;
  }
  assert( obj->nxt );
  oidx = obj->nxt->sym->idx;
  if( !obj->nxt->nxt ) {    
    /* emit CALL.OBJ  oidx, idx */    
    emit_byte(0x06); emit_byte(oidx); emit_byte(idx);    
  }
  else {
    /* emit CALL.OBJ  oidx[*], idx */
    emit_expr(obj->nxt->nxt);
    emit_byte(0x07); emit_byte(oidx); emit_byte(idx);
  }
}

static void emit_lookup(Node *n)
{
  int end, mark;
  int opvalue = GetSpinOpcode(n->op);  /* 0x10 LOOKUPVAL, 0x11 LOOKDOWNVAL */
  Node* item;

  /* output the counter, break address and index */
  emit_push_int( (n->op==TK_LOOKUP||n->op==TK_LOOKDOWN) ? 1 : 0 );
  end = emit_addr(NEWFWD);
  assert(n->sub);
  emit_expr(n->sub->sub);
  
  /* output the lookup/down list items */
  item = n->sub->nxt;
  while( item ) {
    if( item->sub->op==TK_RANGE ) {
      emit_expr( item->sub->sub );
      emit_expr( item->sub->nxt );
      emit_byte( opvalue + 0x02 );
    }
    else {
      emit_expr( item->sub );
      emit_byte( opvalue );
    }
    item = item->nxt;
  }
  emit_byte( 0x0f ); /* LOOKDONE */
  resolve(end);  
}

/* This routine sets up the initialization of a new PNUT Spin
 * interpreter. Most of the work is done by a subroutine that is
 * called by the RUNSPIN opcode, 0x15. In brief, it pops the
 * parameters for the initial subroutine from the current stack and
 * pushes these on the new stack, and then sets up the current stack
 * with the arguments for doing COGINIT(-1, @pnut, @parblock).
 * @pnut is the address of the PNUT interpreter (0xf004)
 * @parblock is the address of a PAR block of 5 words (three
 * longs) located in the new stack space; see the PNUT source for
 * the layout of this PAR block.
 * The bottom of the new stack is further initialised to contain a
 * fake return frame, that causes a COGSTOP(COGID) when the new interpreter
 * returns or aborts from its main subroutine.
 */
static void setup_pnut(Node *n)
{
  Node *arg = n->sub->nxt;
  Symbol *sym;

  /* push the arguments to the method parameter */
  emit_args( arg->nxt );
  
  /* push a long with argc<<8 + idx */  
  sym = (Symbol*)(arg->sub);
  emit_byte(0x39); emit_byte(sym->argc); emit_byte(sym->idx);
  
  /* push the third argument: address of the new stack */
  arg = n->sub->nxt->nxt;
  assert( arg );
  emit_expr( arg->sub );
  
  /* now emit RUNSPIN to call the setup subroutine */
  emit_byte(0x15);
  
  /* if the first argument is !=-1, fixup the stack; this
   * first argument is at LONG[DCURR][-4].
   */
  arg = n->sub->sub;
  if( arg->op!=TK_INTLITERAL || arg->val!=-1 ) {
    emit_expr( arg );
    emit_byte(0x3f); emit_byte(0x8f);  /* FETCH.L   REG[DCURR] */
    emit_byte(0x37); emit_byte(0x61);  /* PUSH      #-4        */
    emit_byte(0xd1);                  /* STORE.L   MEM[*][*]  */
  }
  /* a call to COGINIT (0x28 or 0x2c) follows */
}
