
/* Hive Spin Compiler PUB and PRI block parsing */
/* This code is public domain */

#include "keyword.h"
#include "symbol.h"
#include "parse.h"
#include "hsc.h"
#include <stdio.h>
#include <stdlib.h>
#include <assert.h>

/* Parse a PUB or PRI block. In pass 1 we just count the number
 * of arguments to the method and assign it an index number; also
 * scan the body and allocate space in the DAT section for inline
 * strings.
 */
void ParsePUBPRIpass1()
{
  int symtyp, argc = 0;
  Symbol *sym;
   
  /* create the symbol and assign its index ordinal */
  symtyp = (token.type==TK_PUB) ? PUB_SYMBOL : PRI_SYMBOL;
  NextToken();
  if( !IsId(token.type) ) hsc_abort("expected identifier");
  sym = AddSymbol(token.text, &(curobj->Globals));
  sym->type = symtyp;
  sym->idx = (symtyp==PUB_SYMBOL) ? curobj->nPubs++ : curobj->nPris++;
  NextToken();
   
  /* count the number of parameters */
  if( TestAndSkip(TK_LPAREN) ) {
    do {
      ++argc;
      if( !IsId(token.type) ) hsc_abort("expected identifier");
      NextToken();
    } while( TestAndSkip(TK_COMMA) );
    MustSkip(TK_RPAREN);
  }
  if( argc>255 ) hsc_abort("too many parameters");
  sym->argc = argc;
   
  /* scan the method body for STRING("...") and allocate space in the
   * DAT section; ignore everything else.
   */
  do {
    NextToken();
    if( token.type==TK_STRING ) {
      NextToken();
      MustSkip(TK_LPAREN);
      do {
        if( !IsIntLiteral(token.type) ) hsc_abort("syntax error");
        curobj->datptr++;
        NextToken();
      } while( TestAndSkip(TK_COMMA) );
      MustSkip(TK_RPAREN);
      curobj->datptr++; /* trailing zero byte */
    }
  }
  while( !IsBlockDesignator(token.type) && token.type!=TK_EOF );
}

/* In pass 2 we do more work: the entire header line is parsed and all
 * parameterd and locals are assigned their stack indexes. Note that the
 * stack grows up and that the base pointer is at the start of the
 * actual paramenters, i.e. both parameters and locals have positive
 * offsets. All locals are of type LONG.
 */
void ParsePUBPRIpass2()
{
  Symbol *sym, *method;
  int dim, bp_offset = 4;
  
  LocalsInit();
  NextToken();
  method = SymbolLookup(token.text, curobj->Globals);
  assert( method );
  NextToken();
  
  /* bp+0 = RESULT, bp+4 is first param, etc. */
  if( TestAndSkip(TK_LPAREN) ) {
    do {
      sym = AddSymbol(token.text, &Locals);
      sym->type = LCL_SYMBOL;
      sym->size = 4;
      sym->offset = bp_offset;
      bp_offset += 4;
      NextToken();
    } while( TestAndSkip(TK_COMMA) );
    MustSkip(TK_RPAREN);
  }
  
  /* parse return value and create its symbol */  
  if( TestAndSkip(TK_COLON) ) {
    if( !token.type==TK_ID && !token.type==TK_RESULT ) {
      hsc_abort("expected identifier");
    }
    sym = AddSymbol(token.text, &Locals);
    NextToken();
  }
  else {
    sym = AddSymbol("RESULT", &Locals);
  }
  sym->type = LCL_SYMBOL;
  sym->size = 4;
  sym->offset = 0;
  
  /* parse local variables and add the symbols */
  if( TestAndSkip(TK_BAR) ) {
    if( token.type!=TK_ID ) hsc_abort("expected identifier");
    do {
      sym = AddSymbol(token.text, &Locals);
      sym->type = LCL_SYMBOL;
      sym->size = 4;
      sym->offset = bp_offset;
      NextToken();
      dim = 1;
      if( TestAndSkip(TK_LBRACKET) ) {
        dim = EvalExpr( NULL );
        NextToken();
      }
      bp_offset += dim << 2;
    } while( TestAndSkip(TK_COMMA) );
  }
  MustSkip(TK_EOL);
  method->size = bp_offset;
  
  printf("\nMethod %s\n", method->name);
  PrintSymbols(Locals);
  
  EmitMethodHdr(method);
  CompileBody();
  EmitMethodTail(method);

  LocalsFree();
}
