
/* Hive Spin Compiler statement parsing */
/* This code is public domain */

#include "keyword.h"
#include "symbol.h"
#include "parse.h"
#include "hsc.h"
#include <string.h>

static char pool[64*1024];
static char *brk;
static char *end;

Node* new_node(int op, Node *sub)
{
  Node *n = (Node*)brk;
  brk += sizeof(Node);
  if( brk>=end ) hsc_abort("internal error: parse table full");
  n->op  = op;
  n->sub = sub;
  n->nxt = NULL;
  return n;
}

Node* new_node2(int op, Node *sub, Node *nxt)
{
  Node *n = new_node(op, sub);
  n->nxt = nxt;
  return n;
}

static Node* parse_stmts(int indent);
static Node* parse_simple();
static Node* parse_if();
static Node* parse_case();
static Node* parse_repeat();
static Node* parse_qnxt();

// FIXME
#include <stdio.h>

void CompileBody()
{
  Node *n;
  
  brk = pool;
  end = brk + sizeof(pool);
  n = parse_stmts(-1);
  EmitBlock(n);
  printf("\n");  
}

/* Parse a list of statements. The list is ordered with the first
 * element representing the first statement. Each 'sub' of the list
 * represents one top-level statement of the block being parsed.
 */
static Node* parse_stmts(int indent)
{
  Node *s, *b = NULL, *e = NULL;
  int t = token.type;
  
  while( !IsBlockDesignator(t) && t!=TK_EOF && token.col>indent ) {
    if( TestAndSkip(TK_EOL) ) continue;
    switch( t ) {
      case TK_ABORT:
      case TK_RETURN:     s = parse_simple();   break;
      case TK_IF:
      case TK_IFNOT:      s = parse_if(t);      break;
      case TK_REPEAT:     s = parse_repeat();   break;
      case TK_CASE:       s = parse_case();     break;
      case TK_NEXT:
      case TK_QUIT:       s = parse_qnxt();     break;
      default: 
        s = ParseVoidExpr();
        MustSkip(TK_EOL);
    }
    if( s ) {
      /* add s to the end, e, of list b */
      s = new_node(TK_LIST, s);
      e = e ? (e->nxt = s) : s;
      if( !b ) b = e;
    }
    t = token.type;
  }
  return b;
}

/* This routine handles RETURN and ABORT, possibly followed by
 * an expression. It returns a single node, with TK_RETURN or
 * TK_ABORT in 'op' and the expression, if any, in 'sub'.
 */
static Node* parse_simple()
{
  Node *n = new_node(token.type, NULL);
  NextToken();
  if( token.type!=TK_EOL ) n->sub = ParseExpr();
  MustSkip(TK_EOL);
  return n;
}

/* This routine handles IF, IFNOT, etc. If there is only a THEN
 * clause then it is one node, with TK_IF or TK_IFNOT in 'op' and
 * the THEN clause in 'nxt'. If there is an ELSE clause 'nxt' points
 * to a second node with op 'TK_ELSE' and the THEN clause in 'sub'
 * and the ELSE clause in 'nxt'. Note that ELSEIF and ELSEIFNOT are
 * handled as an ELSE clause with an IF/IFNOT sub-statement.
 */
static Node* parse_if()
{
  Node *expr, *then, *elze, *nxt;
  int t = token.type, col = token.col;
  
  NextToken();
  expr = ParseExpr();
  MustSkip(TK_EOL);
  then = parse_stmts(col);
  if( token.col==col ) {
    if( token.type==TK_ELSEIF || token.type==TK_ELSEIFNOT ) {
      token.type = (token.type==TK_ELSEIF) ? TK_IF : TK_IFNOT;
      elze = parse_if();
      nxt = new_node2(TK_ELSE, then, elze);
      return new_node2(t, expr, nxt);
    }
    else if( TestAndSkip(TK_ELSE) ) {
      elze = parse_stmts(col);
      nxt = new_node2(TK_ELSE, then, elze);
      return new_node2(t, expr, nxt);
    }
    else {
      return new_node2(t, expr, then);
    }
  }
  return new_node2(t, expr, then);
}

/* Allocate a new case item and link it to the end of the list */
static CaseItem* add_case(CaseItem *prev)
{
  CaseItem *ci = (CaseItem*)brk;
  brk += sizeof(CaseItem);
  if( brk>=end ) hsc_abort("internal error: parse table full");
  memset(ci, 0, sizeof(CaseItem));
  if( prev ) prev->next = ci;
  return ci;
}

/* A case statement is returned as a regular node, with 'op' set to
 * TK_CASE, and the case expression in 'sub'. The 'nxt' field points
 * to a linked list of case items, one for each case expression (values
 * go in 'le1', ranges in 'le1'..'le2'). The accompanying statements are
 * pointed to by 'block'. Multiple case expressions for a single block
 * (i.e. comma separated) each get a case item structure, but only the last
 * has a non-null block pointer.
 */
static Node* parse_case()
{
  int itemcol, casecol = token.col;
  Node *top;
  CaseItem *ci = NULL;
  
  NextToken();
  top = new_node(TK_CASE, ParseExpr());
  MustSkip(TK_EOL);
  
  itemcol = token.col;
  if( itemcol<=casecol ) hsc_abort("no case items found");

  while( token.col>casecol ) {
    /* OTHER gets its own case item; both le1 and le2 are null.
     * An OTHER clause ends the CASE statement.
     */
    if( TestAndSkip(TK_OTHER) ) {
      ci = add_case(ci);
      MustSkip(TK_COLON);
      ci->op = TK_OTHER;
      ci->body = parse_stmts( itemcol );
      break;
    }
    else {
      /* Add regular cases */
      do {
        ci = add_case(ci);
        if( !top->nxt ) top->nxt = (Node*)ci;
        ci->op = TK_CASE;
        ci->le1 = ParseExpr();
        if( TestAndSkip(TK_DOTDOT) ) {
          ci->le2 = ParseExpr();
        }
      } while( TestAndSkip(TK_COMMA) );
      MustSkip(TK_COLON);
      ci->body = parse_stmts(itemcol);
    }
  }
  return top;
}

static RepeatBlock *current_repeat = NULL;

/* QUIT and NEXT are parsed into a single node with only 'op' set. The parser
 * checks for the presence of an enclosing REPEAT statement, so that the code
 * generator does not have to.
 */
static Node* parse_qnxt()
{
  Node *n;
  if( !current_repeat ) hsc_abort("QUIT or NEXT outside of REPEAT statement");
  n = new_node(token.type, NULL);
  NextToken();
  MustSkip(TK_EOL);
  return n;
}

/* Allocate a new repeat block and link it to the enclosing repeat block (if any) */
static RepeatBlock* new_repeat()
{
  RepeatBlock *rpt = (RepeatBlock*)brk;
  brk += sizeof(RepeatBlock);
  if( brk>=end ) hsc_abort("internal error: parse table full");
  memset(rpt, 0, sizeof(RepeatBlock));
  rpt->enclosing = current_repeat;
  current_repeat = rpt;
  return rpt;
}

#define IsVar(t) (t==DAT_SYMBOL||t==VAR_SYMBOL||t==LCL_SYMBOL)

/* Only assignable ('lval') expressions can appear before a FROM */ 
static int is_valid_var(Node *n)
{
  int t = n->op;
  return  (IsVar(t) || IsSize(t) || IsReg(t) || t==TK_SPR) ? 1 : 0;
}

/* There are 7 syntactic forms of REPEAT:
 * 1:  REPEAT <stmts>
 * 2:  REPEAT <stmts> WHILE <expr>
 * 2:  REPEAT <stmts> UNTIL <expr>
 * 4:  REPEAT WHILE <expr> <stmts>
 * 5:  REPEAT UNTIL <expr> <stmts>
 * 6:  REPEAT <expr> <stmts>
 * 7:  REPEAT <var> FROM <expr> TO <expr> [STEP <expr>] <stmts>
 * All these forms get stored in a RepeatBlock structure, with the 'op'
 * field identifying the syntactic form.
 */
static Node* parse_repeat()
{
  RepeatBlock *rpt = new_repeat();
  int rptcol = token.col;

  NextToken();
  /* the first form we test is REPEAT on a line by itself; this can
   * develop into syntax 1, 2 or 3.
   */
  if( TestAndSkip(TK_EOL) ) {
    rpt->body = parse_stmts(rptcol);
    rpt->op = TK_RPT;
    if( token.col==rptcol ) {
      if( token.type==TK_WHILE || token.type==TK_UNTIL ) {
        rpt->op = (token.type==TK_WHILE) ? TK_RPTWHILE : TK_RPTUNTIL;
        NextToken();
        rpt->e = ParseExpr();
        MustSkip(TK_EOL);
      }
    }
  }
  /* next we test for form 4 and 5 */
  else if( token.type==TK_WHILE || token.type==TK_UNTIL ) {
    rpt->op = token.type;
    NextToken();
    rpt->e = ParseExpr();
    MustSkip(TK_EOL);
    rpt->body = parse_stmts(rptcol);
  }
  /* and finally for form 6 and 7 */
  else {
    rpt->e = ParseExpr();
    rpt->op = TK_RPTCNT;
    if( is_valid_var(rpt->e) && TestAndSkip(TK_FROM) ) {
      rpt->op = TK_RPTFROM;
      rpt->from = ParseExpr();
      MustSkip(TK_TO);
      rpt->to = ParseExpr();
      if( TestAndSkip(TK_STEP) ) {
        rpt->step = ParseExpr();
      }
    }
    MustSkip(TK_EOL);
    rpt->body = parse_stmts(rptcol);
  }
  current_repeat = rpt->enclosing;
  return (Node*)rpt;
}
