
/* Hive Spin Compiler symbol table */
/* This code is public domain */

#include "symbol.h"
#include <string.h>
#include <stdio.h>
#include <stdlib.h>

/* The local symbol table for PUB/PRI blocks */
Symbol *Locals;

static Symbol *st_brk, *st_max;

/* Find string str in symbol table st.
 * Returns pointer to the symbol node if str is
 * found, NULL otherwise
 */
Symbol *SymbolLookup(char *str, Symbol *st)
{
  int i;
  
  if( !st ) return NULL;
  i = strcmp(str, st->name);
  if( !i ) return st;
  return SymbolLookup(str, i<0 ? st->left : st->right);
}

/* Add string str to the symbol table st.
 * Returns a pointer to the new symbol node
 */
Symbol *AddSymbol(char *str, Symbol **st)
{
  int i;
  
  Symbol *p = *st;
  if( !p ) {
    /* not found, add a new node and link it back into bt */
    *st = p = st_brk++;
    if( st_brk>=st_max ) hsc_abort("symbol table overflow");
    strcpy(p->name, str);
    p->left = NULL;
    p->right = NULL;
    return p;
  }
  i = strcmp(str, p->name);
  if( !i ) hsc_abort("identifier not unique");
  return AddSymbol(str, i<0 ? &p->left : &p->right);
}

static void add_builtin(char *name, int type, int value, Symbol **st)
{
  Symbol *tmp = AddSymbol(name, st);
  tmp->value = value;
  tmp->type = type;
}

Symbol *AllocSymbol(char *name)
{
  Symbol *sym;
  sym = st_brk++;
  strcpy(sym->name, name);
  return sym;
}

DATSym *MakeDATSym(Symbol *sym)
{
  if( sym!=(st_brk-1) ) hsc_abort("internal error: DATSym allocation");
  return (DATSym*)(st_brk++);
}
         
/* Start simple with a set maximum of 1000 symbols. This needs
 * to be improved.
 */
Symbol *SymInit()
{
  Symbol *root = NULL;
  st_brk = malloc( sizeof(Symbol) * 1000);
  st_max = st_brk + 1000;

  /* use a root node about halfway up the alphabet */
  add_builtin( "NEGX",     INT_CON_SYMBOL,    0x80000000, &root );
  
  add_builtin( "FALSE",    INT_CON_SYMBOL,    0,    &root );
  add_builtin( "TRUE",     INT_CON_SYMBOL,    -1,   &root );
  add_builtin( "POSX",     INT_CON_SYMBOL,    0x7fffffff, &root );
  add_builtin( "PI",       FLT_CON_SYMBOL,    0x40490fdb, &root );
  add_builtin( "RCFAST",   INT_CON_SYMBOL,    1,    &root );
  add_builtin( "RCSLOW",   INT_CON_SYMBOL,    2,    &root );
  add_builtin( "XINPUT",   INT_CON_SYMBOL,    4,    &root );
  add_builtin( "XTAL1",    INT_CON_SYMBOL,    8,    &root );
  add_builtin( "XTAL2",    INT_CON_SYMBOL,    16,   &root );
  add_builtin( "XTAL3",    INT_CON_SYMBOL,    32,   &root );
  add_builtin( "PLL1X",    INT_CON_SYMBOL,    64,   &root );
  add_builtin( "PLL2X",    INT_CON_SYMBOL,    128,  &root );
  add_builtin( "PLL4X",    INT_CON_SYMBOL,    256,  &root );
  add_builtin( "PLL8X",    INT_CON_SYMBOL,    512,  &root );
  add_builtin( "PLL16X",   INT_CON_SYMBOL,    1024, &root );
  return root;
}

static Symbol *st_mark;

void LocalsInit()
{
  Locals = NULL;
  st_mark = st_brk;
}

void LocalsFree()
{
  st_brk = st_mark;
}

/* For debug purposes: */

static void print_sym(Symbol *st)
{
  DATSym *dat = (DATSym*)(st+1);
  
  if( !st ) return;
  print_sym( st->left );
  
  switch( st->type ) {
    case INT_CON_SYMBOL:
      printf("%-6s ", "INTCON");
      printf("0x%08x ", st->value);
      printf("%s\n", st->name);
      break;
    case FLT_CON_SYMBOL:
      printf("%-6s ", "FLTCON");
      printf("0x%08x ", st->value);
      printf("%s\n", st->name);
      break;
    case VAR_SYMBOL:
      printf("%-6s ", "VAR");
      printf("           ");
      printf("%s ", st->name);
      printf("(hubloc:%04x, size:%d)\n", st->offset, st->size);
      break;
    case LCL_SYMBOL:
      printf("%-6s ", "LCL");
      printf("           ");
      printf("%s ", st->name);
      printf("(bp_loc:%04x, size:%d)\n", st->offset, st->size);
      break;
    case PUB_SYMBOL:
      printf("%-6s ", "PUB");
      printf("           ");
      printf("%s ", st->name);
      printf("(idx:%3d, argc:%3d)\n", st->idx, st->argc);
      break;
    case PRI_SYMBOL:
      printf("%-6s ", "PRI");
      printf("           ");
      printf("%s ", st->name);
      printf("(idx:%3d, argc:%3d)\n", st->idx, st->argc);
      break;
    case DAT_SYMBOL:
      printf("%-6s ", "DAT");
      printf("0x%08x ", dat->hubloc);
      printf("%s ", st->name);
      printf("(cogloc:%03x, size:%d)\n", dat->cogloc>>2, st->size);
      break;    
  }

  print_sym( st->right );
}

void PrintSymbols(Symbol *st)
{
  printf("Type   Value      Name\n");
  print_sym(st);
}